import logging
from collections import deque
from typing import Optional

from nicegui import ui

from .. import rosys
from ..event import Event
from .communication import Communication
from .lizard_firmware import LizardFirmware

CLOCK_OFFSET_HISTORY_LENGTH = 100


class RobotBrain:
    """This module manages the communication with a [Zauberzeug Robot Brain](https://zauberzeug.com/products/robot-brain).

    It expects a communication object, which is used for the actual read and write operations.
    Besides providing some basic methods like configuring or restarting the microcontroller, it augments and verifies checksums for each message.

    It also keeps track of the clock offset between the microcontroller and the host system, which is used to synchronize the hardware time with the system time.
    The clock offset is calculated by comparing the hardware time with the system time and averaging the differences over a number of samples.
    If the offset changes significantly, a notification is sent and the offset history is cleared.
    """

    def __init__(self, communication: Communication) -> None:
        self.LINE_RECEIVED = Event()
        """a line has been received from the microcontroller (argument: line as string)"""

        self.log = logging.getLogger('rosys.robot_rain')

        self.communication = communication
        self.lizard_code = ''
        self.lizard_firmware = LizardFirmware(self)

        self.waiting_list: dict[str, Optional[str]] = {}
        self._clock_offset: Optional[float] = None
        self._clock_offsets: deque[float] = deque(maxlen=CLOCK_OFFSET_HISTORY_LENGTH)
        self.hardware_time: Optional[float] = None

        rosys.on_startup(self.enable_esp)

    @property
    def clock_offset(self) -> Optional[float]:
        return self._clock_offset

    def developer_ui(self) -> None:
        async def online_update() -> None:
            await self.lizard_firmware.download()
            await self.lizard_firmware.flash_core()
            await self.restart()
            await self.configure()
            await self.lizard_firmware.flash_p0()

        async def local_update() -> None:
            await self.lizard_firmware.flash_core()
            await self.restart()
            await self.configure()
            await self.lizard_firmware.flash_p0()

        with ui.row().classes('items-center'):
            ui.label().bind_text_from(self.lizard_firmware, 'online_version', backward=lambda x: f'Online: {x or "?"}')
            online_update_button = ui.button(on_click=online_update).props('icon=file_download flat round dense') \
                .tooltip('Download and flash online version to Core and P0 microcontrollers')
        with ui.row().classes('items-center'):
            ui.label().bind_text_from(self.lizard_firmware, 'local_version', backward=lambda x: f'Local: {x or "?"}')
            local_update_button = ui.button(on_click=local_update).props('icon=file_download flat round dense') \
                .tooltip('Flash local version to Core and P0 microcontrollers')
        with ui.row().classes('items-center'):
            ui.label().bind_text_from(self.lizard_firmware, 'core_version', backward=lambda x: f'Core: {x or "?"}')
            configure_button = ui.button(on_click=self.configure).props('icon=build flat round dense') \
                .tooltip('Configure microcontrollers')
        with ui.row().classes('items-center'):
            ui.label().bind_text_from(self.lizard_firmware, 'p0_version', backward=lambda x: f'P0: {x or "?"}')

        def update_visibility() -> None:
            online_update_button.visible = \
                self.lizard_firmware.online_version != self.lizard_firmware.core_version or \
                self.lizard_firmware.online_version != self.lizard_firmware.p0_version
            local_update_button.visible = \
                self.lizard_firmware.local_version != self.lizard_firmware.core_version or \
                self.lizard_firmware.local_version != self.lizard_firmware.p0_version
            configure_button.visible = \
                self.lizard_firmware.local_checksum != self.lizard_firmware.core_checksum
        ui.timer(1.0, update_visibility)

        with ui.row().classes('items-center'):
            ui.button('Refresh', on_click=self.lizard_firmware.read_all).props('outline') \
                .tooltip('Read installed and available versions')
            with ui.row():
                with ui.menu() as menu:
                    ui.menu_item('Download', on_click=self.lizard_firmware.download) \
                        .tooltip('Download the latest Lizard firmware from GitHub')
                    ui.menu_item('Flash Core', on_click=self.lizard_firmware.flash_core) \
                        .tooltip('Flash the downloaded Lizard firmware to the Core microcontroller')
                    ui.menu_item('Flash P0', on_click=self.lizard_firmware.flash_p0) \
                        .tooltip('Flash the downloaded Lizard firmware to the P0 microcontroller')
                    ui.menu_item('Enable', on_click=self.enable_esp) \
                        .tooltip('Enable the microcontroller module (will later be done automatically)')
                    ui.menu_item('Configure', on_click=self.configure) \
                        .tooltip('Configure the microcontroller with the Lizard startup file')
                    ui.menu_item('Restart', on_click=self.restart) \
                        .tooltip('Restart the microcontroller')
                ui.button(on_click=menu.open).props('icon=more_vert flat round')

        ui.label().bind_text_from(self, 'clock_offset', lambda offset: f'Clock offset: {offset or 0:.3f} s')

    async def configure(self) -> None:
        rosys.notify('Configuring Lizard...')
        await self.send('!-')
        for line in self.lizard_code.splitlines():
            await self.send(f'!+{line}')
        await self.send('!.')
        await self.restart()
        rosys.notify('Lizard configured successfully.', 'positive')

    async def restart(self) -> None:
        await self.send('core.restart()')

    async def read_lines(self) -> list[tuple[float, str]]:
        lines: list[tuple[float, str]] = []
        millis = None
        while True:
            unchecked = await self.communication.read()
            line = check(unchecked)
            if not line:
                break
            words = line.split()
            if not words:
                continue
            first = words.pop(0)
            if first in self.waiting_list:
                self.waiting_list[first] = line
            if first == 'core':
                millis = float(words.pop(0))
                if self.clock_offset is None:
                    continue
                self.hardware_time = millis / 1000 + self.clock_offset
            self.LINE_RECEIVED.emit(line)
            if self.hardware_time is None:
                continue
            lines.append((self.hardware_time, line))
        if millis is not None:
            self._handle_clock_offset(rosys.time() - millis / 1000)
        return lines

    def _handle_clock_offset(self, offset: float) -> None:
        if self._clock_offset is not None and abs(offset - self._clock_offset) > 0.1:
            self.log.info(f'Clock offset changed from {self._clock_offset:.3f} to {offset:.3f}')
            self._clock_offsets.clear()
        self._clock_offsets.append(offset)
        self._clock_offset = sum(self._clock_offsets) / len(self._clock_offsets)

    async def send(self, msg: str) -> None:
        await self.communication.send(augment(msg))

    async def send_and_await(self, msg: str, ack: str, *, timeout: float = float('inf')) -> Optional[str]:
        self.waiting_list[ack] = None
        await self.send(msg)
        t0 = rosys.time()
        while self.waiting_list.get(ack) is None and rosys.time() < t0 + timeout:
            await rosys.sleep(0.1)
        return self.waiting_list.pop(ack) if ack in self.waiting_list else None

    async def enable_esp(self) -> None:
        rosys.notify('Enabling ESP...')
        command = ['sudo', './flash.py'] + self.lizard_firmware.flash_params + ['enable']
        output = await rosys.run.sh(command, timeout=None, working_dir=self.lizard_firmware.PATH)
        self.log.debug(output)
        rosys.notify('Enabling ESP: done', 'positive')

    def __del__(self) -> None:
        self.communication.disconnect()

    def __repr__(self) -> str:
        return f'<RobotBrain {self.communication}>'


def augment(line: str) -> str:
    checksum = 0
    for c in line:
        checksum ^= ord(c)
    return f'{line}@{checksum:02x}'


def check(line: Optional[str]) -> str:
    if line is None:
        return ''
    if line[-3:-2] != '@':
        return ''
    check_ = int(line[-2:], 16)
    line = line[:-3]
    checksum = 0
    for c in line:
        checksum ^= ord(c)
    if checksum != check_:
        return ''
    return line



import logging
import subprocess
from pathlib import Path
from typing import TYPE_CHECKING, Optional

import requests

from .. import rosys
from ..run import awaitable
from .communication import SerialCommunication

if TYPE_CHECKING:
    from rosys.hardware.robot_brain import RobotBrain


class LizardFirmware:
    GITHUB_URL = 'https://api.github.com/repos/zauberzeug/lizard/releases/latest'
    PATH = Path('~/.lizard').expanduser()

    def __init__(self, robot_brain: 'RobotBrain') -> None:
        self.log = logging.getLogger('rosys.lizard_firmware')
        self.robot_brain = robot_brain

        self.flash_params: list[str] = []

        self.core_version: Optional[str] = None
        self.p0_version: Optional[str] = None
        self.local_version: Optional[str] = None
        self.online_version: Optional[str] = None

        self.local_checksum: Optional[str] = None
        self.core_checksum: Optional[str] = None

    async def read_all(self) -> None:
        await self.read_online_version()
        self.read_local_version()
        await self.read_core_version()
        await self.read_p0_version()
        self.read_local_checksum()
        await self.read_core_checksum()

    async def read_online_version(self) -> None:
        response: dict[str, str] = (await rosys.run.io_bound(requests.get, self.GITHUB_URL)).json()
        try:
            self.online_version = response['tag_name'].removeprefix('v')
        except KeyError:
            rosys.notify(response.get('message', 'Could not access online version'), 'warning')

    def read_local_version(self) -> None:
        path = self.PATH / 'build' / 'lizard.bin'
        with path.open('rb') as f:
            head = f.read(150).decode('utf-8', 'backslashreplace')
        self.local_version = head.split(' ')[3].replace('\x00', '').split('lizard')[0].removeprefix('v')

    async def read_core_version(self) -> None:
        deadline = rosys.time() + 5.0
        while rosys.time() < deadline:
            if response := await self.robot_brain.send_and_await('core.version()', 'version:', timeout=1):
                self.core_version = response.split()[-1].split('-')[0][1:]
                return
            self.log.warning('Could not read Lizard version from Core')

    async def read_p0_version(self) -> None:
        deadline = rosys.time() + 5.0
        while rosys.time() < deadline:
            if response := await self.robot_brain.send_and_await('p0.version()', 'p0:', timeout=1):
                self.p0_version = response.split()[-1].split('-')[0][1:]
                return
            self.log.warning('Could not read Lizard version from P0')

    def read_local_checksum(self) -> None:
        checksum = sum(ord(c) for c in self.robot_brain.lizard_code) % 0x10000
        self.local_checksum = f'{checksum:04x}'
        self.log.info(f'local checksum: {self.local_checksum}')

    async def read_core_checksum(self) -> None:
        deadline = rosys.time() + 5.0
        while rosys.time() < deadline:
            if response := await self.robot_brain.send_and_await('core.startup_checksum()', 'checksum:', timeout=1):
                self.core_checksum = response.split()[-1]
                self.log.info(f'core checksum: {self.core_checksum}')
                return
            self.log.warning('Could not read startup checksum from Core')

    @awaitable
    def download(self) -> None:
        url = requests.get(self.GITHUB_URL, timeout=5.0).json()['assets'][0]['browser_download_url']
        zip_path = self.PATH / 'lizard.zip'
        zip_path.write_bytes(requests.get(url, timeout=5.0).content)
        subprocess.run(['unzip', '-o', zip_path], cwd=self.PATH, check=True)
        zip_path.unlink()
        self.read_local_version()

    async def flash_core(self) -> None:
        assert isinstance(self.robot_brain.communication, SerialCommunication)
        rosys.notify(f'Flashing Lizard firmware {self.local_version} to Core...')
        self.robot_brain.communication.disconnect()
        await rosys.sleep(0.3)
        output = await rosys.run.sh(['sudo', './flash.py'] + self.flash_params, timeout=None, working_dir=self.PATH)
        self.log.info(f'flashed Lizard:\n {output}')
        self.robot_brain.communication.connect()
        await self.read_core_version()
        rosys.notify('Finished.', 'positive')

    async def flash_p0(self) -> None:
        rosys.notify(f'Flashing Lizard firmware {self.core_version} to P0...')
        await self.robot_brain.send('p0.flash()')
        start = rosys.time()
        while (self.robot_brain.hardware_time or 0) < start + 3.0:
            if rosys.time() > start + 60.0:
                rosys.notify('Failed.', 'negative')
                return
            await rosys.sleep(0.1)
        await self.robot_brain.restart()
        await rosys.sleep(3.0)
        await self.read_p0_version()
        rosys.notify('Finished.', 'positive')



from __future__ import annotations

from dataclasses import dataclass

from .point import Point


@dataclass(slots=True, kw_only=True)
class Rectangle:
    x: float
    y: float
    width: float
    height: float

    @property
    def tuple(self) -> tuple[float, float, float, float]:
        return (self.x, self.y, self.width, self.height)

    def contains(self, point: Point) -> bool:
        return (self.x <= point.x <= self.x + self.width and
                self.y <= point.y <= self.y + self.height)

    def __str__(self) -> str:
        return f'Rectangle({round(self.x, 2)}, {round(self.y, 2)}, {round(self.width, 2)}, {round(self.height, 2)})'



import asyncio
import logging
import os
import shlex
import signal
import subprocess
import uuid
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor
from concurrent.futures.process import BrokenProcessPool
from contextlib import contextmanager
from functools import partial, wraps
from pathlib import Path
from typing import Any, Callable, Generator, Optional

from psutil import Popen

from .helpers import is_stopping, is_test

process_pool = ProcessPoolExecutor()
thread_pool = ThreadPoolExecutor(thread_name_prefix='run.py thread_pool')
running_cpu_bound_processes: list[str] = []  # NOTE is used in tests to advance time slower until computation is done
running_sh_processes: list[Popen] = []
log = logging.getLogger('rosys.run')


async def io_bound(callback: Callable, *args: Any, **kwargs: Any):
    if is_stopping():
        return
    loop = asyncio.get_running_loop()
    try:
        return await loop.run_in_executor(thread_pool, partial(callback, *args, **kwargs))
    except RuntimeError as e:
        if 'cannot schedule new futures after shutdown' not in str(e):
            raise
    except asyncio.exceptions.CancelledError:
        pass


def awaitable(func: Callable) -> Callable:
    """decorator to wrap a normal function into an asyncio coroutine"""
    @wraps(func)
    async def inner(*args, **kwargs):
        return await io_bound(func, *args, **kwargs)
    return inner


async def cpu_bound(callback: Callable, *args: Any):
    if is_stopping():
        return
    with cpu():
        try:
            loop = asyncio.get_running_loop()
            return await loop.run_in_executor(process_pool, callback, *args)
        except BrokenProcessPool:
            pass
        except RuntimeError as e:
            if 'cannot schedule new futures after shutdown' not in str(e):
                raise
        except asyncio.exceptions.CancelledError:
            pass


@contextmanager
def cpu() -> Generator[None, None, None]:
    id_ = str(uuid.uuid4())
    running_cpu_bound_processes.append(id_)
    try:
        yield
    finally:
        running_cpu_bound_processes.remove(id_)


async def sh(command: list[str] | str, *,
             timeout: Optional[float] = 1,
             shell: bool = False,
             working_dir: Optional[Path] = None) -> str:
    """executes a shell command

    Args:
        command: a sequence of program arguments as subprocess.Popen requires or full string
        shell: whether a subshell should be launched (default is `False`, for speed, use `True` if you need file globbing or other features)

    Returns:
        stdout
    """
    def popen() -> str:
        cmd_list = command if isinstance(command, list) else shlex.split(command)
        if timeout is not None:
            cmd_list = ['timeout', '--signal=SIGTERM', str(timeout)] + cmd_list
        cmd = ' '.join(cmd_list) if shell else cmd_list
        try:
            with subprocess.Popen(
                cmd,
                cwd=working_dir,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                shell=shell,
                start_new_session=True,
            ) as proc:
                running_sh_processes.append(proc)
                stdout, stderr = proc.communicate()
                assert proc.stdout is not None
                assert proc.stderr is not None
                proc.stdout.close()
                proc.stderr.close()
                _kill(proc)
                running_sh_processes.remove(proc)
                return stdout.decode('utf-8') if proc.returncode == 0 else stderr.decode('utf-8')
        except Exception:
            log.exception(f'failed to run command "{cmd}"')
            if 'proc' in locals() and proc:
                _kill(proc)
            return ''

    if is_stopping():
        return ''
    try:
        return await asyncio.wait_for(io_bound(popen), timeout) or ''
    except asyncio.TimeoutError:
        log.warning(f'Command "{command}" timed out after {timeout} seconds.')
        return ''


def _kill(proc: Popen) -> None:
    try:
        os.killpg(os.getpgid(proc.pid), signal.SIGTERM)
        log.info(f'sent SIGTERM to {proc.pid}')
        try:
            proc.wait(timeout=5)  # wait for 5 seconds
        except subprocess.TimeoutExpired:
            os.killpg(os.getpgid(proc.pid), signal.SIGKILL)  # force kill if process didn't terminate
            log.info(f'sent SIGKILL to {proc.pid}')
            proc.wait()  # ensure the process is reaped
    except ProcessLookupError:
        pass
    except Exception:
        log.exception(f'Failed to kill and/or wait for process {proc.pid}')


def tear_down() -> None:
    # stopping process as described in https://www.cloudcity.io/blog/2019/02/27/things-i-wish-they-told-me-about-multiprocessing-in-python/
    log.info('teardown thread_pool...')
    thread_pool.shutdown(wait=False, cancel_futures=True)
    for process in running_sh_processes:
        _kill(process)
    running_sh_processes.clear()
    if not is_test():
        log.info('teardown process_pool...')
        for process in process_pool._processes.values():  # pylint: disable=protected-access
            process.kill()
        process_pool.shutdown(wait=True, cancel_futures=True)
    log.info('teardown complete.')



import uuid
from enum import Enum
from typing import Optional

import numpy as np
from nicegui import binding, ui
from nicegui.events import SceneClickEventArguments, SceneDragEventArguments

from .. import rosys
from ..event import Event
from ..geometry import Point
from .area import Area
from .path_planner import PathPlanner


class AreaManipulationMode(Enum):
    IDLE = None
    EDIT = 'Edit'
    DELETE = 'Delete'


class Translator:
    Edit = 'Edit'
    Delete = 'Delete'
    Cancel = 'Cancel'
    Done = 'Done'
    Remove_last_point = 'Remove last point'
    Edges_must_not_intersect = 'Edges must not intersect'
    Areas_must_have_at_least_3_points = 'Areas must have at least 3 points'


class AreaManipulation:
    mode = binding.BindableProperty(on_change=lambda sender, value: sender.MODE_CHANGED.emit(value))
    area_type = binding.BindableProperty()
    area_color = binding.BindableProperty()

    def __init__(self, path_planner: PathPlanner) -> None:
        self.path_planner = path_planner
        self.active_area: Optional[Area] = None
        self.mode = AreaManipulationMode.IDLE
        self.area_type = None
        self.area_color = 'green'
        self.translator: Translator = rosys.translator or Translator()

        self.MODE_CHANGED = Event()
        """the mode has changed (argument: new mode)"""

    def create_ui(self) -> ui.row:
        t = self.translator
        with ui.row() as row:
            area_toggle = ui.toggle({
                AreaManipulationMode.EDIT: t.Edit,
                AreaManipulationMode.DELETE: t.Delete,
            }).props('outline').bind_value(self, 'mode')
            ui.button(icon='close', on_click=self.cancel) \
                .props('outline') \
                .tooltip(t.Cancel) \
                .bind_visibility_from(area_toggle, 'value', value=AreaManipulationMode.EDIT)
            ui.button(icon='done', on_click=self.done) \
                .props('outline') \
                .tooltip(t.Done) \
                .bind_visibility_from(area_toggle, 'value', bool)
            ui.button(icon='undo', on_click=self.undo) \
                .props('outline') \
                .tooltip(t.Remove_last_point) \
                .bind_visibility_from(self, 'can_undo')
        return row

    @property
    def can_undo(self) -> bool:
        return self.active_area is not None and len(self.active_area.outline) > 0

    def add_point(self, point: Point) -> None:
        if self.active_area is None:
            area = Area(id=str(uuid.uuid4()), outline=[], closed=False, type=self.area_type, color=self.area_color)
            self.path_planner.areas[area.id] = area
            self.active_area = area

        if self.active_area.would_cause_self_intersection(point):
            rosys.notify(self.translator.Edges_must_not_intersect, type='negative')
            return

        self.active_area.outline.append(point)
        self._emit_change_event(self.active_area)

    def delete_area(self, point: Point) -> None:
        for area in self.path_planner.areas.values():
            if area.contains(point):
                self.path_planner.areas.pop(area.id)
                self.active_area = None
                self._emit_change_event(area)
                return
        if not self.path_planner.areas:
            self.mode = AreaManipulationMode.IDLE

    def delete_point(self, area: Area, point_index: int) -> None:
        del area.outline[point_index]
        if not area.outline:
            self.path_planner.areas.pop(area.id)
            self.active_area = None
        self._emit_change_event(area)

    def undo(self) -> None:
        if not self.can_undo:
            return
        assert self.active_area is not None
        self.active_area.outline.pop()
        if not self.active_area.outline:
            self.path_planner.areas.pop(self.active_area.id)
            self.active_area = None
        self._emit_change_event(self.active_area)

    def cancel(self) -> None:
        if self.active_area:
            self.path_planner.areas.pop(self.active_area.id)
            self._emit_change_event(self.active_area)
        self.active_area = None
        self.mode = AreaManipulationMode.IDLE

    def try_close_active_area(self) -> bool:
        if self.active_area:
            if len(self.active_area.outline) < 3:
                rosys.notify(self.translator.Areas_must_have_at_least_3_points, type='negative')
                return False
            if self.active_area.would_cause_self_intersection(self.active_area.outline[0]):
                rosys.notify(self.translator.Edges_must_not_intersect, type='negative')
                return False
            self.active_area.closed = True
            self._emit_change_event(self.active_area)
        self.active_area = None
        return True

    def done(self) -> None:
        if self.try_close_active_area():
            self.mode = AreaManipulationMode.IDLE

    def handle_click(self, e: SceneClickEventArguments) -> None:
        if e.click_type == 'dblclick':
            for hit in e.hits:
                target = Point(x=hit.x, y=hit.y)
                if hit.object_id == 'ground':
                    if self.mode == AreaManipulationMode.EDIT:
                        self.add_point(target)
                    elif self.mode == AreaManipulationMode.DELETE:
                        self.delete_area(target)
                    return
                if (hit.object_name or '').startswith('area_') and '_corner_' in hit.object_name:
                    area_id = hit.object_name.split('_')[1]
                    area = self.path_planner.areas[area_id]
                    distances = [point.distance(target) for point in area.outline]
                    point_index = int(np.argmin(distances))
                    if area is self.active_area and point_index == 0:
                        self.try_close_active_area()
                    else:
                        self.delete_point(area, point_index)
                    return

    def handle_drag_end(self, e: SceneDragEventArguments) -> None:
        _, area_id, type_, point_index_str = e.object_name.split('_')
        area = self.path_planner.areas[area_id]
        point_index = int(point_index_str)
        new_point = Point(x=e.x, y=e.y)
        if type_ == 'corner':
            old_point = area.outline.pop(point_index)
            if area.would_cause_self_intersection(new_point, point_index):
                rosys.notify(self.translator.Edges_must_not_intersect, type='negative')
                area.outline.insert(point_index, old_point)
            else:
                area.outline.insert(point_index, new_point)
        elif type_ == 'mid':
            new_index = point_index + 1
            if area.would_cause_self_intersection(new_point, new_index):
                rosys.notify(self.translator.Edges_must_not_intersect, type='negative')
            else:
                area.outline.insert(new_index, new_point)
        self._emit_change_event(None)  # TODO: pass area when https://github.com/zauberzeug/nicegui/issues/1505 is fixed

    def _emit_change_event(self, area: Optional[Area] = None) -> None:
        self.path_planner.AREAS_CHANGED.emit([area] if area else None)
        self.path_planner.request_backup()



from .area import Area
from .area_manipulation import AreaManipulation, AreaManipulationMode
from .area_object_ import AreaObject as area_object
from .obstacle import Obstacle
from .obstacle_object_ import ObstacleObject as obstacle_object
from .path_object_ import PathObject as path_object
from .path_planner import PathPlanner

__all__ = [
    'Area',
    'AreaManipulation',
    'AreaManipulationMode',
    'area_object',
    'Obstacle',
    'obstacle_object',
    'path_object',
    'PathPlanner',
]



from dataclasses import fields
from typing import Any

from dataclasses_json.core import _asdict, _decode_dataclass


def to_dict(obj: Any) -> dict[str, Any]:
    """Convert an object `obj` to a serializable dict."""
    return _asdict(obj, False)


def from_dict(cls: type, d: dict[str, Any]) -> Any:
    """Convert a serializable dict `d` to object of type `cls`."""
    return _decode_dataclass(cls, d, False)


def replace_dict(old_dict: dict[str, Any], cls: type, new_dict: dict[str, Any]) -> None:
    """Replace content of `old_dict` with keys and values from `new_dict`."""
    old_dict.clear()
    old_dict.update({key: from_dict(cls, value) for key, value in new_dict.items()})


def replace_list(old_list: list[Any], cls: type, new_list: list[Any]) -> None:
    """Replace content of `old_list` with items from `new_list`."""
    old_list.clear()
    old_list.extend(from_dict(cls, value) for value in new_list)


def replace_set(old_set: set[Any], cls: type, new_set: set[Any]) -> None:
    """Replace content of `old_set` with items from `new_set`."""
    old_set.clear()
    old_set.update(from_dict(cls, value) for value in new_set)


def replace_dataclass(old_dataclass: Any, new_dict: dict[str, Any]) -> None:
    """Replace content of `old_dataclass` with content from `new_dict`."""
    for field in fields(old_dataclass):
        setattr(old_dataclass, field.name, new_dict.get(field.name))



from nicegui import ui

from .steerer import Steerer


class Joystick(ui.joystick):
    """The Joystick UI element allows controlling a given steerer via touch events."""

    def __init__(self, steerer: Steerer, **options) -> None:
        self.steerer = steerer
        super().__init__(on_start=self.handle_start, on_move=self.handle_move, on_end=self.handle_end, **options)

    def handle_start(self, _) -> None:
        self.steerer.start()

    def handle_move(self, msg) -> None:
        self.steerer.update(msg.x, msg.y)

    def handle_end(self, _) -> None:
        self.steerer.stop()



import abc
import logging
from enum import Enum
from typing import Optional
from uuid import uuid4

from ..event import Event
from .detections import Detections
from .image import Image
from .uploads import Uploads


class Autoupload(Enum):
    """Configures the auto-submitting of images to the Learning Loop"""

    FILTERED = 'filtered'
    """only submit images with novel detections and in an uncertainty range (this is the default)"""

    DISABLED = 'disabled'
    """no auto-submitting"""

    ALL = 'all'
    """submit all images which are run through the detector"""


class Detector(abc.ABC):
    """A detector allows detecting objects in images.

    It also holds an upload queue for sending images with uncertain results to an active learning infrastructure like the [Zauberzeug Learning Loop](https://zauberzeug.com/products/learning-loop).
    """

    def __init__(self, *, name: Optional[str] = None) -> None:
        self.name = name or str(uuid4())
        self.NEW_DETECTIONS = Event()
        """detection on an image is completed (argument: image)"""
        self.log = logging.getLogger('rosys.detector')
        self.uploads = Uploads()

    @abc.abstractmethod
    async def detect(self,
                     image: Image,
                     autoupload: Autoupload = Autoupload.FILTERED,
                     tags: list[str] = [],
                     ) -> Detections | None:
        """Runs detections on the image. Afterwards the `image.detections` property is filled."""

    @abc.abstractmethod
    async def upload(self, image: Image, *, tags: list[str] = []) -> None:
        """Uploads the image to the Learning Loop. 

        The `tags` are added to the image.
        If the image has detections, they are also uploaded.
        """



from __future__ import annotations

from dataclasses import dataclass

import numpy as np

from .point import Point


def skew(a, b, c):
    return np.array([[0, -c, b], [c, 0, -a], [-b, a, 0]])


@dataclass(slots=True, kw_only=True)
class Line:
    a: float
    b: float
    c: float

    @property
    def tuple(self) -> tuple[float, float, float]:
        return (self.a, self.b, self.c)

    @property
    def yaw(self) -> float:
        return np.arctan2(self.b, self.a) - np.pi / 2

    @staticmethod
    def from_points(point1: Point, point2: Point) -> Line:
        l = skew(point1.x, point1.y, 1) @ [point2.x, point2.y, 1]
        return Line(a=l[0], b=l[1], c=l[2])

    def intersect(self, other: Line) -> Point:
        p = np.cross(self.tuple, other.tuple)
        return Point(x=p[0] / p[-1], y=p[1] / p[-1])

    def foot_point(self, point: Point) -> Point:
        G3 = np.diag([1, 1, 0])
        p = skew(self.a, self.b, self.c) @ skew(point.x, point.y, 1) @ G3 @ self.tuple
        return Point(x=p[0] / p[-1], y=p[1] / p[-1])

    def distance(self, point: Point) -> float:
        return abs((self.a * point.x + self.b * point.y + self.c) / np.sqrt(self.a**2 + self.b**2))



import random

from ... import persistence, rosys
from ..camera_provider import CameraProvider
from .simulated_camera import SimulatedCamera


class SimulatedCameraProvider(CameraProvider[SimulatedCamera], persistence.PersistentModule):
    """This module collects and simulates cameras and generates synthetic images.

    In the current implementation the images only contain the camera ID and the current time.
    """

    def __init__(self, simulate_failing: bool = False) -> None:
        super().__init__()

        self.simulate_device_failure = simulate_failing

        rosys.on_repeat(self.simulate_device_discovery, 5.0)

    def backup(self) -> dict:
        cameras = {}
        for camera in self._cameras.values():
            cameras[camera.id] = camera.to_dict()
        return {}

    def restore(self, data: dict[str, dict]) -> None:
        for camera_data in data.get('cameras', {}).values():
            camera = SimulatedCamera.from_dict(camera_data)
            self.add_camera(camera)

        for camera in self._cameras.values():
            camera.NEW_IMAGE.register(self.NEW_IMAGE.emit)

    async def simulate_device_discovery(self) -> None:
        for camera in self._cameras.values():
            if not camera.is_connected:
                await camera.connect()
                continue
            assert camera.device is not None

            if self.simulate_device_failure:
                # disconnect cameras randomly with probability rising with time
                time_since_last_activation = rosys.time() - camera.device.creation_time
                if random.random() < time_since_last_activation / 30.0:
                    camera.device = None

    def add_cameras(self, num_cameras: int) -> None:
        for _ in range(num_cameras):
            new_id = f'cam{len(self._cameras)}'
            print(f'adding simulated camera: {new_id}')
            self.add_camera(SimulatedCamera(id=new_id, width=640, height=480))



import asyncio
import logging
import socket
from pathlib import Path

from nicegui import ui

log = logging.getLogger('rosys.wifi')

IP_ADDRESS_LIST = [
    '1.1.1.1',  # Cloudflare
    '1.0.0.1',
    '8.8.8.8',  # Google DNS
    '8.8.4.4',
    '208.67.222.222',  # Open DNS
    '208.67.220.220',
]
PORT = 53


def has_internet() -> bool:
    """Returns True if there's a connection"""
    for host in IP_ADDRESS_LIST:
        try:
            socket.setdefaulttimeout(3)
            socket.socket(socket.AF_INET, socket.SOCK_STREAM).connect((host, PORT))
            return True
        except socket.error:
            pass
    log.warning('No internet connection')
    return False


class WifiButton(ui.button):
    """The WiFi button indicates the current connectivity state and allows setting a new WiFi connection."""

    def __init__(self) -> None:
        with ui.dialog() as self.dialog, ui.card():
            self.status = ui.label()
            self.ssid = ui.input('ssid')
            self.password = ui.input('password')
            with ui.row():
                ui.button('Add', on_click=self.add)
                ui.button('Close', on_click=self.dialog.close)

        super().__init__(on_click=self.dialog.open)
        self.props('icon=network_wifi')
        ui.timer(5, callback=self.update_wifi_status)

    def add(self) -> None:
        log.info(f'adding {self.ssid.value}, {self.password.value}')
        # NOTE a daemon on the host system must watch the .rosys/wifi dir and reload the configuration with nmcli or similar
        config = Path('~/.rosys/wifi').expanduser() / self.ssid.value
        config.mkdir(parents=True, exist_ok=True)
        config.write_text(self.password.value)
        self.dialog.close()

    async def update_wifi_status(self) -> None:
        if await asyncio.get_event_loop().run_in_executor(None, has_internet):
            self.props('color=green')
            self.status.set_text('Robot is connected to the internet.')
        else:
            self.props('color=red')
            self.status.set_text('Robot is not connected to the internet.')



from typing import Optional

from .. import rosys
from .expander import ExpanderHardware
from .module import ModuleHardware
from .robot_brain import RobotBrain


class BatteryControlHardware(ModuleHardware):

    def __init__(self, robot_brain: RobotBrain, *,
                 expander: Optional[ExpanderHardware] = None,
                 name: str = 'battery',
                 reset_pin: int = 15,
                 status_pin: int = 13,
                 ) -> None:
        self.name = name
        self.status: bool = False
        lizard_code = f'{self.name}_reset = {expander.name + "." if expander else ""}Output({reset_pin})\n'
        lizard_code += f'{self.name}_status = {expander.name + "." if expander else ""}Input({status_pin})\n'
        core_message_fields = [f'{self.name}_status.level']
        super().__init__(robot_brain=robot_brain, lizard_code=lizard_code, core_message_fields=core_message_fields)

    def handle_core_output(self, _: float, words: list[str]) -> None:
        self.status = int(words.pop(0)) == 1

    async def release_battery_relay(self) -> None:
        self.log.info('releasing battery relay')
        await rosys.sleep(1)
        await self.robot_brain.send(f'{self.name}_reset.on()')
        await rosys.sleep(0.5)
        await self.robot_brain.send(f'{self.name}_reset.off()')



from .calibratable_camera import CalibratableCamera
from .camera import Camera
from .configurable_camera import ConfigurableCamera
from .transformable_camera import TransformableCamera

__all__ = [
    'CalibratableCamera',
    'Camera',
    'ConfigurableCamera',
    'TransformableCamera',
]



import io
from typing import Any, Optional

import cv2
import imgsize
import numpy as np
import PIL

from ..geometry import Rectangle
from .image import ImageSize
from .image_rotation import ImageRotation


class PeekableBytesIO(io.BytesIO):

    def peek(self, n=-1):
        position = self.tell()
        data = self.read(n)
        self.seek(position)
        return data


def get_image_size_from_bytes(image: bytes) -> ImageSize:
    try:
        with PeekableBytesIO(image) as f:
            width, height = imgsize.get_size(f)
    except imgsize.UnknownSize as e:
        raise ValueError('Could not determine image size') from e
    return ImageSize(width=width, height=height)


def process_jpeg_image(data: bytes, rotation: ImageRotation, crop: Optional[Rectangle] = None) -> bytes:
    """Rotate and crop a JPEG image."""
    image = PIL.Image.open(io.BytesIO(data))
    if crop is not None:
        image = image.crop((int(crop.x), int(crop.y), int(crop.x + crop.width), int(crop.y + crop.height)))
    image = image.rotate(int(rotation), expand=True)  # NOTE: PIL handles rotation with 90 degree steps efficiently
    img_byte_arr = io.BytesIO()
    image.save(img_byte_arr, format='JPEG')
    return img_byte_arr.getvalue()


def process_ndarray_image(image: np.ndarray, rotation: ImageRotation, crop: Optional[Rectangle] = None) -> bytes:
    """Rotate and crop a NumPy image."""
    if crop is not None:
        image = image[int(crop.y):int(crop.y+crop.height), int(crop.x):int(crop.x+crop.width)]
    if rotation == ImageRotation.LEFT:
        image = cv2.rotate(image, cv2.ROTATE_90_COUNTERCLOCKWISE)
    elif rotation == ImageRotation.RIGHT:
        image = cv2.rotate(image, cv2.ROTATE_90_CLOCKWISE)
    elif rotation == ImageRotation.UPSIDE_DOWN:
        image = cv2.rotate(image, cv2.ROTATE_180)
    return cv2.imencode('.jpg', image)[1].tobytes()


def to_bytes(image: Any) -> bytes:
    # TODO: type hint and docstring
    return image[0].tobytes()


def remove_exif(image_data: bytes) -> bytes:  # written by ChatGPT
    pos = 2  # Skip SOI marker
    while pos < len(image_data):
        if image_data[pos] == 0xFF:
            if image_data[pos + 1] == 0xE1:  # APP1 marker (EXIF)
                length = image_data[pos + 2] << 8 | image_data[pos + 3]
                image_data = image_data[:pos] + image_data[pos + length + 2:]
                continue  # Check for multiple EXIF segments
            elif image_data[pos + 1] == 0xD9:  # EOI marker
                break  # End of image
            else:
                length = image_data[pos + 2] << 8 | image_data[pos + 3]
                pos += length + 2  # Skip to next marker
        else:
            pos += 1  # Increment position if not a marker
    return image_data



import logging
import shutil

from ... import persistence, rosys
from ..camera_provider import CameraProvider
from .usb_camera import UsbCamera
from .usb_camera_scanner import scan_for_connected_devices

SCAN_INTERVAL = 10


class UsbCameraProvider(CameraProvider[UsbCamera], persistence.PersistentModule):
    """This module collects and provides real USB cameras.

    Camera devices are discovered through video4linux (v4l) and accessed with openCV.
    Therefore the program v4l2ctl and openCV (including python bindings) must be available.
    """

    def __init__(self) -> None:
        super().__init__()

        self.log = logging.getLogger('rosys.usb_camera_provider')

        rosys.on_shutdown(self.shutdown)
        rosys.on_repeat(self.update_device_list, SCAN_INTERVAL)

    def backup(self) -> dict:
        return {
            'cameras': {camera.id: camera.to_dict() for camera in self._cameras.values()}
        }

    def restore(self, data: dict[str, dict]) -> None:
        for camera_data in data.get('cameras', {}).values():
            self.add_camera(UsbCamera.from_dict(camera_data))

    @staticmethod
    async def scan_for_cameras() -> set[str]:
        return (await rosys.run.io_bound(scan_for_connected_devices)) or set()

    async def update_device_list(self) -> None:
        camera_uids = await self.scan_for_cameras()
        for uid in camera_uids:
            if uid not in self._cameras:
                self.add_camera(UsbCamera(id=uid))
            await self._cameras[uid].connect()

    async def shutdown(self) -> None:
        for camera in self._cameras.values():
            await camera.disconnect()

    @staticmethod
    def is_operable() -> bool:
        return shutil.which('v4l2-ctl') is not None



from typing import Optional

from ..event import Event
from ..geometry import Rotation
from .module import Module, ModuleHardware, ModuleSimulation
from .robot_brain import RobotBrain


class Imu(Module):

    def __init__(self, offset_rotation: Rotation, **kwargs) -> None:
        super().__init__(**kwargs)
        self.offset_rotation = offset_rotation
        self.rotation: Optional[Rotation] = None
        self.gyro_calibration: float = 0.0

        self.NEW_MEASUREMENT = Event()
        """a new measurement has been received (argument: Tuple[Roll, Pitch, Yaw])"""

    @property
    def roll(self) -> Optional[float]:
        return self.rotation.roll if self.rotation else None

    @property
    def pitch(self) -> Optional[float]:
        return self.rotation.pitch if self.rotation else None

    @property
    def yaw(self) -> Optional[float]:
        return self.rotation.yaw if self.rotation else None

    @property
    def euler(self) -> Optional[tuple[float, float, float]]:
        return self.rotation.euler if self.rotation else None

    def emit_measurement(self) -> None:
        self.NEW_MEASUREMENT.emit(self.euler)


class ImuHardware(Imu, ModuleHardware):

    def __init__(self, robot_brain: RobotBrain, name: str = 'imu', **kwargs) -> None:
        self.name = name
        self.lizard_code = f'{name} = Imu()'
        self.core_message_fields = [
            f'{name}.cal_gyr',
            f'{name}.quat_w:4',
            f'{name}.quat_x:4',
            f'{name}.quat_y:4',
            f'{name}.quat_z:4',
        ]
        super().__init__(robot_brain=robot_brain, lizard_code=self.lizard_code, core_message_fields=self.core_message_fields, **kwargs)

    def handle_core_output(self, time: float, words: list[str]) -> None:
        self.gyro_calibration = float(words.pop(0))
        rotation = Rotation.from_quaternion(
            float(words.pop(0)),
            float(words.pop(0)),
            float(words.pop(0)),
            float(words.pop(0)),
        )

        if self.gyro_calibration < 1.0:
            return

        self.rotation = rotation * self.offset_rotation.T
        self.emit_measurement()


class ImuSimulation(Imu, ModuleSimulation):

    def simulate_measurement(self, rotation: Rotation) -> None:
        self.rotation = rotation * self.offset_rotation.T
        self.emit_measurement()



import asyncio
from datetime import datetime, timedelta
from typing import Any, Optional

import socketio
import socketio.exceptions

from .. import persistence, rosys
from ..helpers import LazyWorker
from .detections import BoxDetection, Detections, PointDetection, SegmentationDetection
from .detector import Autoupload, Detector
from .image import Image


class DetectorHardware(Detector):
    """This detector communicates with a [YOLO detector](https://hub.docker.com/r/zauberzeug/yolov5-detector) via Socket.IO.

    It automatically connects and reconnects, submits and receives detections and sends images that should be uploaded to the [Zauberzeug Learning Loop](https://zauberzeug.com/products/learning-loop).
    """

    def __init__(self, *, port: int = 8004, name: Optional[str] = None) -> None:
        super().__init__(name=name)

        self.sio = socketio.AsyncClient()
        self.lazy_worker = LazyWorker()
        self.port = port
        self.timeout_count = 0

        @self.sio.on('disconnect')
        def on_sio_disconnect() -> None:
            self.log.warning(f'sio disconnect on port {port}')

        @self.sio.on('connect_error')
        def on_sio_connect_error(err) -> None:
            self.log.warning(f'sio connect error on {port}: {err}')

        rosys.on_repeat(self.step, 1.0)

    @property
    def is_connected(self) -> bool:
        return self.sio.connected

    async def step(self) -> None:
        if not self.is_connected:
            self.log.info(f'trying reconnect {self.name}')
            if not await self.connect():
                self.log.exception(f'connection to {self.name} at port {self.port} failed; trying again')
                await rosys.sleep(3.0)
                return

        await self.try_start_one_upload()
        await self.upload_priority_queue()

    async def connect(self) -> bool:
        try:
            url = f'ws://localhost:{self.port}'
            self.log.info(f'connecting to detector at {url}')
            await self.sio.connect(url, socketio_path='/ws/socket.io', wait_timeout=3.0)
            self.log.info('connected successfully')
            return True
        except Exception:
            return False

    async def disconnect(self) -> None:
        await self.sio.disconnect()

    async def try_start_one_upload(self) -> None:
        if datetime.now() < self.uploads.last_upload + timedelta(minutes=self.uploads.minimal_minutes_between_uploads):
            return

        upload_images = self.uploads.get_queued()
        if upload_images:
            rosys.background_tasks.create(self.upload(upload_images[0]), name='upload_image')
            self.uploads.queue.clear()  # old images should not be uploaded later when the robot is inactive
            self.uploads.last_upload = datetime.now()

    async def upload_priority_queue(self) -> None:
        upload_images = self.uploads.get_priority_queued()
        if upload_images:
            async def upload_priority_images():
                for image in upload_images:
                    await self.upload(image)
            rosys.background_tasks.create(upload_priority_images(), name='upload_priority_images')
            self.uploads.priority_queue.clear()

    async def upload(self, image: Image, *, tags: list[str] = []) -> None:
        try:
            self.log.info(f'Upload detections to port {self.port}')
            data_dict: dict[str, Any] = {'image': image.data, 'mac': image.camera_id}
            detections = image.get_detections(self.name)
            if detections is not None:
                detections_dict = detections.to_dict()
                detections_dict['box_detections'] = _box_detections_to_int(detections_dict.pop('boxes'))
                detections_dict['point_detections'] = detections_dict.pop('points')
                detections_dict['segmentation_detections'] = detections_dict.pop('segmentations')
                data_dict['detections'] = detections_dict
            data_dict['tags'] = list(image.tags.union(tags))
            await self.sio.emit('upload', data_dict)

        except Exception:
            self.log.exception(f'could not upload {image.id}')

    async def detect(self,
                     image: Image,
                     autoupload: Autoupload = Autoupload.FILTERED,
                     tags: list[str] = [],
                     ) -> Detections | None:
        return await self.lazy_worker.run(self._detect(image, autoupload, tags))

    async def _detect(self, image: Image, autoupload: Autoupload, tags: list[str]) -> Detections | None:
        if image.is_broken:
            return None
        try:
            result: dict = await self.sio.call('detect', {
                'image': image.data,
                'mac': image.camera_id,
                'autoupload': autoupload.value,
                'tags': tags,
            }, timeout=3)
            if image.is_broken:  # NOTE: image can be marked broken while detection is underway
                return None
            detections = Detections(
                boxes=[persistence.from_dict(BoxDetection, d) for d in result.get('box_detections', [])],
                points=[persistence.from_dict(PointDetection, d) for d in result.get('point_detections', [])],
                segmentations=[
                    persistence.from_dict(SegmentationDetection, d)
                    for d in result.get('segmentation_detections', [])
                ],
            )
            image.set_detections(self.name, detections)
            self.timeout_count = 0
            if image.is_broken:
                return None
            self.NEW_DETECTIONS.emit(image)
            return detections
        except socketio.exceptions.TimeoutError:
            self.log.debug(f'detection for {image.id} on {self.port} took too long')
            self.timeout_count += 1
        except asyncio.exceptions.CancelledError:
            self.log.debug('task has been cancelled')
        except Exception:
            self.log.exception(f'could not detect {image.id}')
        finally:
            if self.timeout_count > 5:
                await self.disconnect()
        return None

    def __str__(self) -> str:
        return f'{type(self).__name__} ({"connected" if self.is_connected else "disconnected"})'


def _box_detections_to_int(detections: list[dict]) -> list[dict]:
    for detection in detections:
        detection['x'] = int(detection['x'])
        detection['y'] = int(detection['y'])
        detection['width'] = int(detection['width'])
        detection['height'] = int(detection['height'])
    return detections



import logging
from asyncio import Task
from asyncio.subprocess import Process
from io import BytesIO
from typing import AsyncGenerator, Optional

import httpx
from nicegui import background_tasks

from ..image_processing import remove_exif
from .vendors import mac_to_url


class MjpegDevice:

    def __init__(self, mac: str, ip: str, *,
                 index: Optional[int] = None, username: Optional[str] = None, password: Optional[str] = None) -> None:
        self.mac = mac
        self.ip = ip
        self.capture_task: Optional[Task] = None
        self.capture_process: Optional[Process] = None
        self._image_buffer: Optional[bytes] = None
        self._authorized: bool = True
        self.authentication = None if username is None or password is None else httpx.DigestAuth(username, password)
        self.log = logging.getLogger('rosys.mjpeg_device ' + self.mac)
        url = mac_to_url(mac, ip, index=index)
        if url is None:
            raise ValueError(f'could not determine URL for {mac}')
        self.url = url

        self.start_capture_task()

    def start_capture_task(self):
        self.capture_task = background_tasks.create(self.run_capture_task(), name=f'capture {self.mac}')

    async def restart_capture(self) -> None:
        self.shutdown()
        self.start_capture_task()

    async def run_capture_task(self) -> None:
        self.log.info(f'Capturing images from {self.url}')

        async def stream() -> AsyncGenerator[bytes, None]:
            async with httpx.AsyncClient() as client:
                assert self.url is not None
                try:
                    async with client.stream('GET', self.url, auth=self.authentication) as response:  # type: ignore
                        if response.status_code != 200:
                            self.log.error(f'could not connect to {self.url} '
                                           f'(credentials: {self.authentication}): '
                                           f'{response.status_code} {response.reason_phrase}')
                            return
                        buffer = BytesIO()
                        header = None
                        pos = 0
                        try:
                            async for chunk in response.aiter_bytes():
                                buffer.write(chunk)
                                while True:
                                    if header is None:
                                        header_pos = buffer.getvalue().find(b'\xff\xd8', pos)
                                        if header_pos == -1:
                                            pos = max(0, buffer.tell() - 1)
                                            break
                                        pos = header_pos + 2
                                        header = header_pos
                                    else:
                                        footer_pos = buffer.getvalue().find(b'\xff\xd9', pos)
                                        if footer_pos == -1:
                                            pos = max(0, buffer.tell() - 1)
                                            break
                                        image_data = buffer.getvalue()[header:footer_pos + 2]
                                        yield remove_exif(image_data)
                                        buffer = BytesIO(buffer.getvalue()[footer_pos + 2:])
                                        pos = 0
                                        header = None
                        except httpx.ReadTimeout:
                            self.log.warning(f'Connection to {self.url} timed out')
                except Exception:
                    self.log.warning(f'Initial connection to {self.url} failed. Was something disconnected?')

        async for image in stream():
            self._image_buffer = image

        self.capture_task = None

    def capture(self) -> Optional[bytes]:
        return self._image_buffer

    def shutdown(self) -> None:
        if self.capture_process is not None:
            self.capture_process.terminate()
            self.capture_process = None



from .app_controls_ import AppButton
from .app_controls_ import AppControls as app_controls
from .automation_controls_ import AutomationControls as automation_controls
from .automator import Automator
from .parallelize import parallelize
from .schedule import Schedule

__all__ = [
    'AppButton',
    'app_controls',
    'automation_controls',
    'Automator',
    'parallelize',
    'Schedule',
]



from dataclasses import dataclass


@dataclass
class KpiChart:
    title: str
    indicators: dict[str, str]
    colormap: str = ''  # choose from matplotlib colormaps
    scale: float = 1.0
    unit: str = ''



import asyncio
from typing import Any, Coroutine, TypeVar

_T = TypeVar('_T')


class LazyWorker:

    def __init__(self) -> None:
        self.waiting: asyncio.Task | None = None
        self.is_free = asyncio.Event()
        self.is_free.set()

    async def run(self, coro: Coroutine[Any, None, _T]) -> _T | None:
        task: asyncio.Task[_T] = asyncio.create_task(coro)
        self.waiting = task
        while True:
            if self.waiting != task:
                task.cancel()
                return None
            try:
                await asyncio.wait_for(self.is_free.wait(), timeout=0.1)
            except asyncio.TimeoutError:
                continue
            if self.waiting == task:
                break

        self.is_free.clear()
        self.waiting = None
        result = await task
        self.is_free.set()
        return result



import logging

from nicegui import ui

LEVELS = {
    logging.NOTSET: 'NOTSET',
    logging.DEBUG: 'DEBUG',
    logging.INFO: 'INFO',
    logging.WARNING: 'WARNING',
    logging.ERROR: 'ERROR',
    logging.CRITICAL: 'CRITICAL',
}


class LoggingPage:
    """Logging Page

    This module creates a page to change the log levels of different loggers.
    A list of logger names like `["field_friend", "rosys"]` can be passed to group them together.
    It is mounted at /logging.
    """

    def __init__(self, group_names: list[str] = []) -> None:

        @ui.page('/logging')
        def page():
            loggers = [logging.getLogger(name) for name in logging.root.manager.loggerDict]  # pylint: disable=no-member
            groups = {name: [l for l in loggers if l.name.startswith(name)] for name in group_names}
            groups['others'] = [l for l in loggers if not any(l.name.startswith(name) for name in group_names)]

            @ui.refreshable
            def _column(group_name: str, loggers: list[logging.Logger]) -> None:
                with ui.column():
                    ui.label(group_name).classes('text-xl')
                    for logger in sorted(loggers, key=lambda logger: logger.name.lower()):
                        ui.select(LEVELS, label=logger.name, value=logger.getEffectiveLevel(),
                                  on_change=lambda e, logger=logger: _update_level(logger, e.value)) \
                            .classes('w-64').props('dense outlined')

            def _update_level(logger: logging.Logger, level: int) -> None:
                logger.setLevel(level)
                _column.refresh()

            with ui.row():
                for name, loggers in groups.items():
                    _column(name, loggers)



import asyncio
import gc
import logging
import multiprocessing
import os
import signal
import threading
import time as pytime
from dataclasses import dataclass
from typing import Any, Awaitable, Callable, Literal, Optional

import numpy as np
import psutil
from nicegui import Client, app, background_tasks, ui

from . import event, persistence, run
from .config import Config
from .helpers import invoke, is_stopping
from .helpers import is_test as is_test_

log = logging.getLogger('rosys.core')

config = Config()
translator: Optional[Any] = None

is_test = is_test_()

# POSIX standard to create processes is "fork", which is inherently broken for python (see https://pythonspeed.com/articles/python-multiprocessing/)
if __name__ == '__main__':
    multiprocessing.set_start_method('spawn')


@dataclass(slots=True, kw_only=True)
class Notification:
    time: float
    message: str


NEW_NOTIFICATION = event.Event()
"""notify the user (string argument: message)"""


class _state:
    start_time: float = 0.0 if is_test else pytime.time()
    time = start_time
    last_time_request: float = start_time
    exception: Optional[BaseException] = None  # NOTE: used for tests


def get_last_exception() -> Optional[BaseException]:
    return _state.exception


notifications: list[Notification] = []
repeat_handlers: list[tuple[Callable, float]] = []
startup_handlers: list[Callable] = []
shutdown_handlers: list[Callable] = []
tasks: list[asyncio.Task] = []


def notify(message: str,
           type: Optional[Literal[  # pylint: disable=redefined-builtin
               'positive',
               'negative',
               'warning',
               'info',
               'ongoing',
           ]] = None) -> None:
    log.info(message)
    notifications.append(Notification(time=time(), message=message))
    NEW_NOTIFICATION.emit(message)
    # NOTE show notifications on all pages
    for client in Client.instances.values():
        if not client.has_socket_connection:
            continue
        with client:
            try:
                ui.notify(message, type=type)
            except Exception:
                log.exception('failed to call notify')


time_lock = threading.Lock()


def time() -> float:
    if is_test:
        return _state.time
    with time_lock:
        now = pytime.time()
        _state.time += (now - _state.last_time_request) * config.simulation_speed
        _state.last_time_request = now
        return _state.time


def set_time(value: float) -> None:
    assert is_test, 'only tests can change the time'
    _state.time = value


def uptime() -> float:
    return time() - _state.start_time


async def sleep(seconds: float) -> None:
    if is_test:
        sleep_end_time = time() + seconds
        while time() <= sleep_end_time:
            await asyncio.sleep(0)
    else:
        if config.simulation_speed <= 0:
            config.simulation_speed = 0.01
        scaled_seconds = seconds / config.simulation_speed
        count = int(np.ceil(scaled_seconds))
        if count > 0:
            for _ in range(count):
                await asyncio.sleep(scaled_seconds / count)
        else:
            await asyncio.sleep(0)


def _run_handler(handler: Callable) -> None:
    try:
        result = handler()
        if isinstance(result, Awaitable):
            tasks.append(background_tasks.create(result, name=handler.__qualname__))
    except Exception:
        log.exception(f'error while starting handler "{handler.__qualname__}"')


def _start_loop(handler: Callable, interval: float) -> None:
    log.debug(f'starting loop "{handler.__qualname__}" with interval {interval:.3f}s')
    tasks.append(background_tasks.create(_repeat_one_handler(handler, interval), name=handler.__qualname__))


def on_repeat(handler: Callable, interval: float) -> None:
    if tasks:  # RoSys is already running
        _start_loop(handler, interval)
    else:
        repeat_handlers.append((handler, interval))


def on_startup(handler: Callable) -> None:
    if tasks:  # RoSys is already running
        _run_handler(handler)
    else:
        startup_handlers.append(handler)


def on_shutdown(handler: Callable) -> None:
    shutdown_handlers.append(handler)


async def startup() -> None:
    if tasks:
        raise RuntimeError('should be only executed once')
    if multiprocessing.get_start_method() != 'spawn':
        raise RuntimeError(
            'multiprocessing start method must be "spawn"; see https://pythonspeed.com/articles/python-multiprocessing/')

    persistence.restore()

    for handler in startup_handlers:
        _run_handler(handler)

    for coroutine in event.startup_coroutines:
        await coroutine

    for handler, interval in repeat_handlers:
        _start_loop(handler, interval)


async def _garbage_collection(mbyte_limit: float = 300) -> None:
    if psutil.virtual_memory().free < mbyte_limit * 1_000_000:
        log.warning(f'less than {mbyte_limit} mb of memory remaining -> start garbage collection')
        gc.collect()
        await sleep(1)  # NOTE ensure all warnings appear before sending "finished" message
        log.warning('finished garbage collection')


async def _watch_emitted_events() -> None:
    try:
        for task in event.tasks:
            if task.done() and task.exception():
                _state.exception = task.exception()
                log.exception('task failed to execute', exc_info=task.exception())
        event.tasks = [t for t in event.tasks if not t.done()]
    except Exception:
        log.exception('failed to watch emitted events')


async def _repeat_one_handler(handler: Callable, interval: float) -> None:
    await sleep(interval)  # NOTE delaying first execution so not all actors rush in at the same time
    while True:
        start = time()
        try:
            if is_stopping():
                log.info(f'{handler} must be stopped')
                break
            await invoke(handler)
            dt = time() - start
        except (asyncio.CancelledError, GeneratorExit):
            return
        except Exception:
            dt = time() - start
            log.exception(f'error in "{handler.__qualname__}"')
            if interval == 0 and dt < 0.1:
                delay = 0.1 - dt
                log.warning(
                    f'"{handler.__qualname__}" would be called to frequently ' +
                    f'because it only took {dt*1000:.0f} ms; ' +
                    f'delaying this step for {delay*1000:.0f} ms')
                await sleep(delay)
        try:
            await sleep(interval - dt)
        except (asyncio.CancelledError, GeneratorExit):
            return


async def shutdown() -> None:
    for handler in shutdown_handlers:
        log.debug(f'invoking shutdown handler "{handler.__qualname__}"')
        await invoke(handler)
    log.debug('creating data backup')
    await persistence.backup(force=True)
    log.debug('tear down "run" tasks')
    run.tear_down()
    log.debug('canceling all remaining tasks')
    for task in tasks:
        task.cancel()
    log.debug('clearing tasks')
    tasks.clear()

    # NOTE: kill own process and all its children after uvicorn has finished (we had many restart freezes before)
    def delayed_kill():
        pytime.sleep(1)
        os.kill(os.getpid(), signal.SIGKILL)
    if not is_test:
        threading.Thread(target=delayed_kill).start()
    log.debug('finished shutdown')


def reset_before_test() -> None:
    assert is_test
    set_time(0)  # NOTE: in tests we start at zero for better readability
    _state.exception = None


def reset_after_test() -> None:
    assert is_test
    startup_handlers.clear()
    repeat_handlers[3:] = []  # NOTE: remove all but internal handlers
    shutdown_handlers.clear()
    event.reset()


gc.disable()  # NOTE disable automatic garbage collection to optimize performance
on_repeat(_garbage_collection, 10 * 60)
on_repeat(_watch_emitted_events, 0.1)
on_repeat(persistence.backup, 10)

app.on_startup(startup)
app.on_shutdown(shutdown)



import logging
from typing import Callable, Coroutine, Optional, cast

from .. import rosys
from ..analysis import track
from ..driving import Steerer
from ..event import Event
from .automation import Automation


class Automator:
    """An automator allows running automations, i.e. coroutines that can be paused and resumed.

    See [Click-and-drive](https://rosys.io/examples/click-and-drive/) for a simple example of an automation.

    _steerer_: If provided, manually steering the robot will pause a currently running automation.

    _default_automation_: If provided, it allows the automator to start a new automation without passing an automation
    (e.g. via an "Play"-button like offered by the [automation controls](https://rosys.io/reference/rosys/automation/#rosys.automation.automation_controls)). 
    The passed function should return a new coroutine on every call (see [Play-pause-stop](https://rosys.io/examples/play-pause-stop/) example).

    _on_interrupt_: Optional callback that will be called when an automation pauses or stops.
    """

    def __init__(self,
                 steerer: Optional[Steerer], *,
                 default_automation: Optional[Callable] = None,
                 on_interrupt: Optional[Callable] = None) -> None:
        self.AUTOMATION_STARTED = Event()
        """an automation has been started"""

        self.AUTOMATION_PAUSED = Event()
        """an automation has been paused (string argument: description of the cause)"""

        self.AUTOMATION_RESUMED = Event()
        """an automation has been resumed"""

        self.AUTOMATION_STOPPED = Event()
        """an automation has been stopped (string argument: description of the cause)"""

        self.AUTOMATION_FAILED = Event()
        """an automation has failed to complete (string argument: description of the cause)"""

        self.AUTOMATION_COMPLETED = Event()
        """an automation has been completed"""

        self.log = logging.getLogger('rosys.automator')

        self.default_automation = default_automation

        self.enabled: bool = True
        self.automation: Optional[Automation] = None

        if steerer:
            steerer.STEERING_STARTED.register(lambda: self.pause(because='steering started'))

        if on_interrupt:
            self.AUTOMATION_PAUSED.register(lambda _: rosys.background_tasks.create(cast(Callable, on_interrupt)()))
            self.AUTOMATION_STOPPED.register(lambda _: rosys.background_tasks.create(cast(Callable, on_interrupt)()))

        rosys.on_shutdown(lambda: self.stop(because='automator is shutting down'))

    @property
    def is_stopped(self) -> bool:
        return self.automation is None or self.automation.is_stopped

    @property
    def is_running(self) -> bool:
        return self.automation is not None and self.automation.is_running

    @property
    def is_paused(self) -> bool:
        return self.automation is not None and self.automation.is_paused

    def start(self, coro: Optional[Coroutine] = None, *, paused: bool = False) -> None:
        """Starts a new automation.

        You can pass any coroutine.
        The automator will make sure it can be paused, resumed and stopped.
        """
        if coro is None:
            assert self.default_automation is not None
            self.start(self.default_automation(), paused=paused)
            return
        if not self.enabled:
            coro.close()
            return
        self.stop(because='new automation starts')
        self.automation = Automation(coro, self._handle_exception, on_complete=self._on_complete)
        rosys.background_tasks.create(self.automation.__await__(), name='automation')  # type: ignore
        self.AUTOMATION_STARTED.emit()
        rosys.notify('automation started')
        if paused:
            self.automation.pause()

    def pause(self, because: str) -> None:
        """Pauses the current automation.

        You need to provide a cause which will be used as notification message.
        """
        if self.is_running:
            assert self.automation is not None
            self.automation.pause()
            self.AUTOMATION_PAUSED.emit(because)
            rosys.notify(f'automation paused because {because}')

    def resume(self) -> None:
        """Resumes the current automation."""
        if not self.enabled:
            return
        if self.is_paused:
            assert self.automation is not None
            self.automation.resume()
            self.AUTOMATION_RESUMED.emit()
            rosys.notify('automation resumed')

    def stop(self, because: str) -> None:
        """Stops the current automation.

        You need to provide a cause which will be used as notification message.
        """
        if not self.is_stopped:
            assert self.automation is not None
            self.automation.stop()
            self.AUTOMATION_STOPPED.emit(because)
            track.reset()
            rosys.notify(f'automation stopped because {because}')

    def enable(self) -> None:
        """Enables the automator.

        It is enabled by default.
        It can be disabled by calling `disable()`.
        """
        self.enabled = True

    def disable(self, because: str) -> None:
        """Disables the automator.

        No automations can be started while the automator is disabled.
        If an automation is running or paused it will be stopped.
        You need to provide a cause which will be used as notification message.
        """
        self.stop(because)
        self.enabled = False

    def _handle_exception(self, e: Exception) -> None:
        self.stop(because='an exception occurred in an automation')
        self.AUTOMATION_FAILED.emit(str(e))
        rosys.notify('automation failed', 'negative')
        if rosys.is_test:
            self.log.exception('automation failed')

    def _on_complete(self) -> None:
        self.AUTOMATION_COMPLETED.emit()
        rosys.notify('automation completed', 'positive')



from contextlib import nullcontext

from nicegui.elements.scene_object3d import Object3D
from nicegui.elements.scene_objects import Curve

from ..driving import PathSegment


class PathObject(Object3D):

    def __init__(self) -> None:
        super().__init__('group')

    def update(self, path: list[PathSegment], height: float = 0) -> None:
        for obj in list(self.scene.objects.values()):
            if obj.name == 'path':
                obj.delete()
        with self.scene or nullcontext():
            for segment in path:
                Curve(
                    [segment.spline.start.x, segment.spline.start.y, height],
                    [segment.spline.control1.x, segment.spline.control1.y, height],
                    [segment.spline.control2.x, segment.spline.control2.y, height],
                    [segment.spline.end.x, segment.spline.end.y, height],
                ).material('#ff8800').with_name('path')



from typing import Optional


class VendorType:
    AXIS = 1
    OTHER = -1


def mac_to_vendor(mac: str) -> int:
    if any(mac.startswith(prefix) for prefix in ['00:40:8c', 'ac:cc:8e', 'b8:a4:4f', 'e8:27:25']):
        return VendorType.AXIS
    return VendorType.OTHER


def mac_to_url(mac: str, ip: str, *, index: Optional[int] = None) -> Optional[str]:
    vendor = mac_to_vendor(mac)
    if vendor == VendorType.AXIS:
        return f'http://{ip}/axis-cgi/mjpg/video.cgi' + (f'?camera={index}' if index is not None else '')
    return None



import logging
from typing import Any, Optional

import cv2
from typing_extensions import Self

from ... import rosys
from ..camera.configurable_camera import ConfigurableCamera
from ..camera.transformable_camera import TransformableCamera
from ..image import Image, ImageSize
from ..image_processing import process_jpeg_image, process_ndarray_image, to_bytes
from ..image_rotation import ImageRotation
from .usb_device import UsbDevice


class UsbCamera(ConfigurableCamera, TransformableCamera):

    def __init__(self,
                 *,
                 id: str,  # pylint: disable=redefined-builtin
                 name: Optional[str] = None,
                 connect_after_init: bool = True,
                 streaming: bool = True,
                 auto_exposure: bool = True,
                 exposure: bool = False,
                 width: int = 800,
                 height: int = 600,
                 fps: int = 10,
                 **kwargs) -> None:
        super().__init__(id=id,
                         name=name,
                         connect_after_init=connect_after_init,
                         streaming=streaming,
                         **kwargs)
        self._pending_operations = 0
        self.device: Optional[UsbDevice] = None
        self.detect: bool = False
        self.color: Optional[str] = None

        self._register_parameter('auto_exposure', self.get_exposure, self.set_exposure, auto_exposure)
        self._register_parameter('exposure', self.get_exposure, self.set_exposure, exposure)
        self._register_parameter('width', self.get_width, self.set_width, width)
        self._register_parameter('height', self.get_height, self.set_height, height)
        self._register_parameter('fps', self.get_fps, self.set_fps, fps)

    def to_dict(self) -> dict[str, Any]:
        return super().to_dict() | {
            name: param.value for name, param in self._parameters.items()
        }

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Self:
        return cls(**data)

    @property
    def is_connected(self) -> bool:
        return self.device is not None

    async def connect(self) -> None:
        if self.is_connected:
            return

        device = UsbDevice.from_uid(self.id)
        if device is None:
            logging.warning(f'Connecting camera {self.id}: failed')
            return

        self.device = device
        logging.info(f'Connecting camera {self.id}: succeeded')

        self._apply_all_parameters()

    async def disconnect(self) -> None:
        if not self.is_connected:
            return

        assert self.device is not None
        await rosys.run.io_bound(self.device.capture.release)
        self.device = None
        logging.info(f'camera {self.id}: disconnected')

    async def capture_image(self) -> None:
        if not self.is_connected:
            return None

        assert self.device is not None

        capture_success, captured_image = await rosys.run.io_bound(self.device.capture.read)
        image_is_MJPG = 'MJPG' in self.device.video_formats

        if not capture_success:
            await self.disconnect()
            return

        if captured_image is None:
            return

        if image_is_MJPG:
            bytes_ = await rosys.run.io_bound(to_bytes, captured_image)
            if self.crop or self.rotation != ImageRotation.NONE:
                bytes_ = await rosys.run.cpu_bound(process_jpeg_image, bytes_, self.rotation, self.crop)
        else:
            bytes_ = await rosys.run.cpu_bound(process_ndarray_image, captured_image, self.rotation, self.crop)

        image_size = ImageSize(width=captured_image.shape[1], height=captured_image.shape[0])
        final_image_resolution = self._resolution_after_transform(image_size)

        image = Image(time=rosys.time(), camera_id=self.id, size=final_image_resolution, data=bytes_)
        self._add_image(image)

    def set_auto_exposure(self, auto: bool) -> None:
        assert self.device is not None

        device = self.device

        if device.has_manual_exposure:
            is_auto_exposure = self.get_auto_exposure()
            if auto and not is_auto_exposure:
                # self.log.info(f'activating auto-exposure for {self.id}')
                device.capture.set(cv2.CAP_PROP_AUTO_EXPOSURE, 3)
            else:
                device.capture.set(cv2.CAP_PROP_AUTO_EXPOSURE, 1)
                self.set_exposure(self._parameters['exposure'].value)

    def set_exposure(self, value: float) -> None:
        assert self.device is not None

        device = self.device
        assert device.capture is not None

        if device.has_manual_exposure:
            is_auto_exposure = self.get_auto_exposure()
            if not is_auto_exposure:
                exposure = device.capture.get(cv2.CAP_PROP_EXPOSURE) / device.exposure_max
                if value != exposure:
                    device.capture.set(cv2.CAP_PROP_EXPOSURE, int(value * device.exposure_max))

    def get_auto_exposure(self) -> Optional[bool]:
        assert self.device is not None
        device = self.device
        return device.capture.get(cv2.CAP_PROP_AUTO_EXPOSURE) == 3

    def get_exposure(self) -> Optional[float]:
        assert self.device is not None

        device = self.device
        if not device.has_manual_exposure:
            return None
        return device.capture.get(cv2.CAP_PROP_EXPOSURE) / device.exposure_max

    def set_width(self, width: int) -> None:
        assert self.device is not None
        device = self.device
        device.capture.set(cv2.CAP_PROP_FRAME_WIDTH, width)

    def get_width(self) -> int:
        assert self.device is not None
        device = self.device
        return int(device.capture.get(cv2.CAP_PROP_FRAME_WIDTH))

    def set_height(self, height: int) -> None:
        assert self.device is not None
        device = self.device
        device.capture.set(cv2.CAP_PROP_FRAME_HEIGHT, height)

    def get_height(self) -> int:
        assert self.device is not None
        device = self.device
        return int(device.capture.get(cv2.CAP_PROP_FRAME_HEIGHT))

    def set_fps(self, fps: int) -> None:
        assert self.device is not None
        device = self.device
        device.capture.set(cv2.CAP_PROP_FPS, fps)

    def get_fps(self) -> int:
        assert self.device is not None
        device = self.device
        return int(device.capture.get(cv2.CAP_PROP_FPS))



import logging
from typing import Optional

import httpx

from ... import persistence, rosys
from ..camera_provider import CameraProvider
from ..rtsp_camera.arp_scan import find_cameras
from .mjpeg_camera import MjpegCamera
from .vendors import VendorType, mac_to_vendor


class MjpegCameraProvider(CameraProvider[MjpegCamera], persistence.PersistentModule):

    def __init__(self, username: Optional[str] = None, password: Optional[str] = None) -> None:
        super().__init__()

        self.username = username
        self.password = password

        self.log = logging.getLogger('rosys.mjpeg_camera_provider')
        rosys.on_shutdown(self.shutdown)
        rosys.on_repeat(self.update_device_list, 5.)

    def restore(self, data: dict[str, dict]) -> None:
        for camera_data in data.get('cameras', {}).values():
            camera_data['password'] = self.password
            camera_data['username'] = self.username
            camera = MjpegCamera.from_dict(camera_data)
            self.add_camera(camera)
        for camera in self._cameras.values():
            camera.NEW_IMAGE.register(self.NEW_IMAGE.emit)

    async def scan_for_cameras(self) -> list[tuple[str, str]]:
        ids_ips: list[tuple[str, str]] = []
        async for mac, ip in find_cameras():
            if mac_to_vendor(mac) != VendorType.AXIS:
                continue

            authentication = None if self.username is None or self.password is None else \
                httpx.DigestAuth(self.username, self.password)
            url = f'http://{ip}/axis-cgi/videostatus.cgi'
            try:
                async with httpx.AsyncClient() as client:
                    response = await client.get(url, auth=authentication)
            except httpx.HTTPError as e:
                self.log.warning(f'Error while looking for cameras at axis router ({url}): {e}')
                continue
            if response.status_code != 200:
                continue

            camera_infos = response.text.split('\n')
            for info_line in camera_infos:
                if info_line.endswith(' = video'):
                    video_name = info_line.removesuffix(' = video')
                    index = video_name.split(' ')[1]
                    ids_ips.append((f'{mac}-{index}', ip))
        return ids_ips

    async def update_device_list(self) -> None:
        newly_disconnected_cameras = {id for id, camera in self._cameras.items() if camera.is_connected}
        for camera_id, ip in await self.scan_for_cameras():
            if camera_id not in self._cameras:
                self.add_camera(MjpegCamera(id=camera_id, username=self.username,
                                password=self.password, connect_after_init=False))
            if camera_id in newly_disconnected_cameras:
                newly_disconnected_cameras.remove(camera_id)
            camera = self._cameras[camera_id]
            if not camera.is_connected:
                self.log.info(f'activating authorized camera "{camera.id}" at ip "{ip}" ...')
                await camera.connect(ip=ip)

        for mac in newly_disconnected_cameras:
            await self._cameras[mac].disconnect()

    async def shutdown(self) -> None:
        for camera in self._cameras.values():
            await camera.disconnect()



from .camera import Camera
from .camera_provider import CameraProvider


class MultiCameraProvider(CameraProvider):
    """A multi-camera provider combines multiple camera providers into one.

    This is useful if another module requires a single camera provider but the robot has multiple camera sources like USB and WiFi cameras.
    """

    def __init__(self, *camera_providers: CameraProvider) -> None:
        super().__init__()
        self.providers = camera_providers

    @property
    def cameras(self) -> dict[str, Camera]:
        return {id: camera for provider in self.providers for id, camera in provider.cameras.items()}



import os
import sys
from typing import AsyncIterator, Optional

from ... import rosys
from .vendors import VendorType, mac_to_vendor


async def run_arp_scan() -> str:
    """Run an arp-scan on the given interface and return the output."""
    if sys.platform.startswith('darwin'):
        arpscan_cmd = 'arp-scan'
    else:
        arpscan_cmd = '/usr/sbin/arp-scan'  # TODO is this necessary? sbin should be in path
    if os.getuid() != 0:
        arpscan_cmd = f'sudo {arpscan_cmd}'
    cmd = f'{arpscan_cmd} --localnet'
    output = await rosys.run.sh(cmd, timeout=10)

    if 'sudo' in output and not 'name resolution' in output:
        raise RuntimeError('Could not run arp-scan! Make sure it is installed and can be run with sudo.'
                           'Try running sudo visudo and add the following line: "rosys ALL=(ALL) NOPASSWD: /usr/sbin/arp-scan"')
    return output


async def find_cameras() -> AsyncIterator[tuple[str, str]]:
    """Find all cameras in the local network and return MAC and IP addresses."""
    output = await run_arp_scan()
    if 'ERROR' in output:
        return
    for line in output.splitlines():
        infos = line.split()
        if len(infos) < 2:
            continue
        mac = infos[1]
        ip = infos[0]
        yield mac, ip


async def find_known_cameras() -> list[str]:
    """Find all cameras of known vendors and return MAC addresses."""
    return [mac async for mac, _ in find_cameras() if mac_to_vendor(mac) != VendorType.OTHER]


async def find_ip(mac: str) -> Optional[str]:
    """Find the IP address of a camera with the given MAC address."""
    async for mac_, ip in find_cameras():
        if mac_ == mac:
            return ip
    return None



import asyncio
import logging
from typing import Any, Callable, Coroutine, Generator, Optional


class Automation:
    """An automation wraps a coroutine and allows pausing and resuming it.

    Optional exception and completion handlers are called if provided.
    This class is used internally by the [automator](https://rosys.io/reference/rosys/automation/#rosys.automation.Automator).
    """

    def __init__(self,
                 coro: Coroutine,
                 exception_handler: Optional[Callable] = None,
                 on_complete: Optional[Callable] = None) -> None:
        self.log = logging.getLogger('rosys.automation')
        self.coro = coro
        self.exception_handler = exception_handler
        self.on_complete = on_complete
        self._can_run = asyncio.Event()
        self._can_run.set()
        self._stop = False
        self._is_waited = False

    @property
    def is_running(self) -> bool:
        return self._is_waited and self._can_run.is_set()

    @property
    def is_paused(self) -> bool:
        return self._is_waited and not self._can_run.is_set()

    @property
    def is_stopped(self) -> bool:
        return not self._is_waited

    def __await__(self) -> Generator[Any, None, Any | None]:
        try:
            self._is_waited = True
            coro_iter = self.coro.__await__()
            iter_send, iter_throw = coro_iter.send, coro_iter.throw
            send: Callable = iter_send
            message: Any = None
            while not self._stop:
                try:
                    while not self._can_run.is_set() and not self._stop:
                        yield from self._can_run.wait().__await__()  # pylint: disable=no-member
                except BaseException as err:
                    send, message = iter_throw, err

                if self._stop:
                    return None

                try:
                    signal = send(message)
                except StopIteration as err:
                    self.log.info('automation is finished')
                    if self.on_complete:
                        self.on_complete()
                    return err.value
                send = iter_send
                try:
                    message = yield signal
                except BaseException as err:
                    send, message = iter_throw, err
        except Exception as e:
            self.log.exception('automation failed')
            if self.exception_handler:
                self.exception_handler(e)
            raise
        finally:
            self._is_waited = False
        return None

    def pause(self) -> None:
        self._can_run.clear()

    def resume(self) -> None:
        self._can_run.set()

    def stop(self) -> None:
        self._can_run.set()
        self._stop = True



import logging
from typing import Optional

from ... import persistence, rosys
from ..camera_provider import CameraProvider
from .arp_scan import find_known_cameras
from .rtsp_camera import RtspCamera

SCAN_INTERVAL = 10


class RtspCameraProvider(CameraProvider[RtspCamera], persistence.PersistentModule):
    """This module collects and provides real RTSP streaming cameras."""

    def __init__(self, *, frame_rate: int = 6, jovision_profile: int = 0) -> None:
        super().__init__()

        self.frame_rate = frame_rate
        self.jovision_profile = jovision_profile

        self.log = logging.getLogger('rosys.rtsp_camera_provider')

        self.last_scan: Optional[float] = None

        rosys.on_shutdown(self.shutdown)
        rosys.on_repeat(self.update_device_list, 10.)

    def backup(self) -> dict:
        cameras = {}
        for camera in self._cameras.values():
            cameras[camera.id] = camera.to_dict()
        return {'cameras': cameras}

    def restore(self, data: dict[str, dict]) -> None:
        for camera_data in data.get('cameras', {}).values():
            camera = RtspCamera.from_dict(camera_data)
            self.add_camera(camera)
        for camera in self._cameras.values():
            camera.NEW_IMAGE.register(self.NEW_IMAGE.emit)

    @staticmethod
    async def scan_for_cameras() -> list[str]:
        return await find_known_cameras()

    async def update_device_list(self) -> None:
        if self.last_scan is not None and rosys.time() < self.last_scan + SCAN_INTERVAL:
            return
        self.last_scan = rosys.time()
        newly_disconnected_cameras = {id for id, camera in self._cameras.items() if camera.is_connected}
        for mac in await find_known_cameras():
            if mac not in self._cameras:
                self.add_camera(RtspCamera(id=mac, fps=self.frame_rate, jovision_profile=self.jovision_profile))
            if mac in newly_disconnected_cameras:
                newly_disconnected_cameras.remove(mac)
            camera = self._cameras[mac]
            if not camera.is_connected:
                self.log.info(f'activating authorized camera {camera.id}...')
                await camera.connect()

        for mac in newly_disconnected_cameras:
            await self._cameras[mac].disconnect()

    async def shutdown(self) -> None:
        for camera in self._cameras.values():
            await camera.disconnect()



import gc
import logging
import os
import tracemalloc
from typing import Optional

import psutil
from fastapi import Request
from psutil._common import bytes2human
from starlette.applications import ASGIApp
from starlette.middleware.base import BaseHTTPMiddleware

from .. import rosys, run

log = logging.getLogger('rosys.analysis.memory')


class MemoryMiddleware(BaseHTTPMiddleware):

    def __init__(self, app: ASGIApp) -> None:
        super().__init__(app)
        self.last_mem: Optional[int] = None

    async def dispatch(self, request: Request, call_next):
        mem = get_process_memory()
        response = await call_next(request)
        if response.headers.get('x-nicegui-content') == 'page' and self.last_mem is not None:
            log.info(f'GET {request.get("path")} increased memory by {bytes2human(mem - self.last_mem)} '
                     f'and is now {bytes2human(mem)}')
        self.last_mem = mem
        return response


def get_process_memory() -> int:
    process = psutil.Process(os.getpid())
    mem_info = process.memory_info()
    return mem_info.rss


def get_humanreadable_process_memory() -> int:
    return bytes2human(get_process_memory())


def compare_tracemalloc_snapshots(snapshot, prev_snapshot):
    stats = snapshot.compare_to(prev_snapshot, 'traceback')
    for stat in stats[:10]:
        trace = '\n'.join(stat.traceback.format())
        usage = f'{bytes2human(stat.size_diff)} new, {bytes2human(stat.size)} total; {stat.count_diff} new memory blocks, {stat.count} total'
        log.info(f'{trace}"\n"{usage}')


def observe_memory_growth(with_tracemalloc: bool = False) -> None:
    log.info('Observing memory growth')
    prev_memory: int = 0
    prev_snapshot: Optional[tracemalloc.Snapshot] = None
    if with_tracemalloc:
        tracemalloc.start(10)

    async def stats() -> None:
        nonlocal prev_memory
        nonlocal prev_snapshot
        gc.collect()
        growth = get_process_memory() - prev_memory
        # log.info('==============')
        log.info(
            f'memory growth: {bytes2human(growth)}, '
            f"now it's {get_humanreadable_process_memory()}"
        )
        # log.info('==============')
        prev_memory = get_process_memory()
        if with_tracemalloc:
            snapshot = tracemalloc.take_snapshot()
            if growth > 4 * 1e-6 and prev_snapshot is not None:
                await run.cpu_bound(compare_tracemalloc_snapshots, snapshot, prev_snapshot)
            prev_snapshot = snapshot

    rosys.on_repeat(stats, 60.0)



import abc

from .. import rosys
from ..event import Event
from ..geometry import Pose, PoseStep, Velocity
from ..helpers import remove_indentation
from .can import CanHardware
from .module import Module, ModuleHardware, ModuleSimulation
from .robot_brain import RobotBrain


class Wheels(Module, abc.ABC):
    """This module represents wheels for a two-wheel differential drive.

    Wheels can be moved using the `drive` methods and provide measured velocities as an event.
    """

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

        self.VELOCITY_MEASURED = Event()
        """new velocity measurements are available for processing (argument: list of velocities)"""

        self.linear_target_speed: float = 0.0
        self.angular_target_speed: float = 0.0

        rosys.on_shutdown(self.stop)

    @abc.abstractmethod
    async def drive(self, linear: float, angular: float) -> None:
        self.linear_target_speed = linear
        self.angular_target_speed = angular

    async def stop(self) -> None:
        await self.drive(0.0, 0.0)


class WheelsHardware(Wheels, ModuleHardware):
    """This module implements wheels hardware.

    Drive and stop commands are forwarded to a given Robot Brain.
    Velocities are read and emitted regularly.
    """

    def __init__(self, robot_brain: RobotBrain, *,
                 can: CanHardware,
                 name: str = 'wheels',
                 left_can_address: int = 0x000,
                 right_can_address: int = 0x100,
                 m_per_tick: float = 0.01,
                 width: float = 0.5,
                 is_left_reversed: bool = False,
                 is_right_reversed: bool = False) -> None:
        self.name = name
        lizard_code = remove_indentation(f'''
            l = ODriveMotor({can.name}, {left_can_address})
            r = ODriveMotor({can.name}, {right_can_address})
            l.m_per_tick = {m_per_tick}
            r.m_per_tick = {m_per_tick}
            l.reversed = {'true' if is_left_reversed else 'false'}
            r.reversed = {'true' if is_right_reversed else 'false'}
            {name} = ODriveWheels(l, r)
            {name}.width = {width}
        ''')
        core_message_fields = [f'{self.name}.linear_speed:3', f'{self.name}.angular_speed:3']
        super().__init__(robot_brain=robot_brain, lizard_code=lizard_code, core_message_fields=core_message_fields)

    async def drive(self, linear: float, angular: float) -> None:
        await super().drive(linear, angular)
        await self.robot_brain.send(f'{self.name}.speed({linear}, {angular})')

    def handle_core_output(self, time: float, words: list[str]) -> None:
        velocity = Velocity(linear=float(words.pop(0)), angular=float(words.pop(0)), time=time)
        self.VELOCITY_MEASURED.emit([velocity])


class WheelsSimulation(Wheels, ModuleSimulation):
    """This module simulates two wheels.

    Drive and stop commands impact internal velocities (linear and angular).
    A simulated pose is regularly updated with these velocities, while the velocities are emitted as an event.
    """

    def __init__(self) -> None:
        super().__init__()

        self.pose: Pose = Pose()
        self.linear_velocity: float = 0
        self.angular_velocity: float = 0
        self.inertia_factor: float = 0.0
        self.friction_factor: float = 0.0
        self.is_blocking: bool = False
        self.is_slipping: bool = False

    async def drive(self, linear: float, angular: float) -> None:
        await super().drive(linear, angular)
        f = self.inertia_factor
        self.linear_velocity = 0 if self.is_blocking else f * self.linear_velocity + (1 - f) * linear
        self.angular_velocity = 0 if self.is_blocking else f * self.angular_velocity + (1 - f) * angular

    async def step(self, dt: float) -> None:
        self.linear_velocity *= 1 - self.friction_factor
        self.angular_velocity *= 1 - self.friction_factor
        if not self.is_slipping:
            self.pose += PoseStep(linear=dt*self.linear_velocity, angular=dt*self.angular_velocity, time=rosys.time())
        velocity = Velocity(linear=self.linear_velocity, angular=self.angular_velocity, time=rosys.time())
        self.VELOCITY_MEASURED.emit([velocity])



import abc
import logging
from typing import Callable

from .robot_brain import RobotBrain


class Module(abc.ABC):

    def __init__(self) -> None:
        super().__init__()
        self.log = logging.getLogger(__name__)


class ModuleHardware(Module):

    def __init__(self, robot_brain: RobotBrain, lizard_code: str, core_message_fields: list[str] = [],
                 **kwargs) -> None:
        super().__init__(**kwargs)
        self.robot_brain = robot_brain
        self.lizard_code = lizard_code
        self.core_message_fields: list[str] = core_message_fields
        self.message_hooks: dict[str, Callable] = {}

    def handle_core_output(self, time: float, words: list[str]) -> None:  # pylint: disable=unused-argument
        pass


class ModuleSimulation(Module):

    async def step(self, dt: float) -> None:  # pylint: disable=unused-argument
        pass



import abc
from typing import Generic, Optional, TypeVar

from .. import rosys
from ..event import Event
from .camera.camera import Camera
from .image import Image

T = TypeVar('T', bound=Camera)


class CameraProvider(Generic[T], rosys.persistence.PersistentModule, metaclass=abc.ABCMeta):
    """A camera provider holds a dictionary of cameras and manages additions and removals.

    The camera dictionary should not be modified directly but by using the camera provider's methods.
    This way respective events are emitted and consistency can be taken care of.

    The camera provider also creates an HTTP route to access camera images.
    """

    def __init__(self) -> None:
        super().__init__()

        self.CAMERA_ADDED = Event()
        """a new camera has been added (argument: camera)"""

        self.CAMERA_REMOVED = Event()
        """a camera has been removed (argument: camera id)"""

        self.NEW_IMAGE = Event()
        """a new image is available (argument: image)"""

        self._cameras: dict[str, T] = {}

    def backup(self) -> dict:
        return {
            'cameras': {id: camera.to_dict() for id, camera in self._cameras.items()},
        }

    def restore(self, data: dict[str, dict]) -> None:
        rosys.persistence.replace_dict(self._cameras, Camera, data.get('cameras', {}))
        for camera in self._cameras.values():
            camera.NEW_IMAGE.register(self.NEW_IMAGE.emit)

    @property
    def cameras(self) -> dict[str, T]:
        return self._cameras

    @property
    def images(self) -> list[Image]:
        return sorted((i for c in self.cameras.values() for i in c.images), key=lambda i: i.time)

    def add_camera(self, camera: T) -> None:
        self._cameras[camera.id] = camera
        self.CAMERA_ADDED.emit(camera)
        camera.NEW_IMAGE.register(self.NEW_IMAGE.emit)
        self.request_backup()

    def remove_camera(self, camera_id: str) -> None:
        del self.cameras[camera_id]
        self.CAMERA_REMOVED.emit(camera_id)
        self.request_backup()

    def remove_all_cameras(self) -> None:
        for camera_id in list(self.cameras):
            self.remove_camera(camera_id)

    def prune_images(self, max_age_seconds: Optional[float] = None):
        for camera in self.cameras.values():
            if max_age_seconds is None:
                camera.images.clear()
            else:
                while camera.images and camera.images[0].time < rosys.time() - max_age_seconds:
                    del camera.images[0]



import asyncio
import logging
from dataclasses import fields, is_dataclass
from typing import Any, Callable, Optional

import numpy as np
import pytest

from .. import rosys, run
from ..automation import Automator
from ..driving import Driver, Odometer
from ..geometry import Point, Point3d

log = logging.getLogger(__name__)

automator: Optional[Automator] = None
driver: Optional[Driver] = None
odometer: Optional[Odometer] = None


async def forward(seconds: Optional[float] = None,
                  *,
                  until: Optional[int | float | Callable] = None,
                  x: Optional[float] = None,
                  y: Optional[float] = None,
                  tolerance: float = 0.1,
                  dt: float = 0.01,
                  timeout: float = 100):
    start_time = rosys.time()
    if seconds is not None:
        def condition():
            return rosys.time() >= start_time + seconds
        msg = f'forwarding {seconds=}'
        timeout = max(timeout, seconds)
    elif isinstance(until, (int, float)):
        def condition():
            return rosys.time() >= until
        msg = f'forwarding {until=}'
        timeout = max(timeout, until - start_time)
    elif callable(until):
        msg = f'forwarding {until=}'
        condition = until
    elif x is not None and y is not None:
        def condition():
            assert odometer is not None
            return (odometer.prediction.x - x)**2 + (odometer.prediction.y - y)**2 < tolerance**2
        msg = f'forwarding to {x=} and {y=}'
    elif x is not None:
        def condition():
            assert odometer is not None
            return abs(odometer.prediction.x - x) < tolerance
        msg = f'forwarding to {x=}'
    elif y is not None:
        def condition():
            assert odometer is not None
            return abs(odometer.prediction.y - y) < tolerance
        msg = f'forwarding to {y=}'
    else:
        raise ValueError('invalid arguments')

    log.info(f'\033[94m{msg}\033[0m')
    while not condition():
        if rosys.time() > start_time + timeout:
            raise TimeoutError(f'condition took more than {timeout} s')
        if not run.running_cpu_bound_processes:
            rosys.set_time(rosys.time() + dt)
            await asyncio.sleep(0)
        else:
            await asyncio.sleep(0.01)
        exception = rosys.get_last_exception()
        if exception is not None:
            raise RuntimeError(f'error while forwarding time {dt} s') from exception


def assert_pose(x: float, y: float, *, deg: Optional[float] = None, position_tolerance: float = 0.1, deg_tolerance: float = 1.0) -> None:
    assert odometer is not None
    assert odometer.prediction.x == pytest.approx(x, abs=position_tolerance)
    assert odometer.prediction.y == pytest.approx(y, abs=position_tolerance)
    if deg is not None:
        assert np.rad2deg(odometer.prediction.yaw) == pytest.approx(deg, abs=deg_tolerance)


def assert_point(actual: Point | Point3d, expected: Point | Point3d, tolerance=0.1) -> None:
    assert type(actual) == type(expected)  # pylint: disable=unidiomatic-typecheck
    assert actual.x == pytest.approx(expected.x, abs=tolerance)
    assert actual.y == pytest.approx(expected.y, abs=tolerance)
    if isinstance(actual, Point3d) and isinstance(expected, Point3d):
        assert actual.z == pytest.approx(expected.z, abs=tolerance)


async def automate_drive_to(x: float, y: float) -> None:
    assert automator is not None
    assert driver is not None
    automator.start(driver.drive_to(Point(x=x, y=y)))


def approx(o1: Any, o2: Any, *,
           rel: Optional[float] = None,
           abs: Optional[float] = None,  # pylint: disable=redefined-builtin
           nan_ok: bool = False) -> None:
    if hasattr(o1, '__dict__'):
        assert sorted(o1.__dict__) == sorted(o2.__dict__)
        for key, value in o1.__dict__.items():
            approx(value, getattr(o2, key), rel=rel, abs=abs, nan_ok=nan_ok)
    elif is_dataclass(o1):
        o1_keys = [f.name for f in fields(o1)]
        o2_keys = [f.name for f in fields(o2)]
        assert sorted(o1_keys) == sorted(o2_keys)
        for key in o1_keys:
            approx(getattr(o1, key), getattr(o2, key), rel=rel, abs=abs, nan_ok=nan_ok)
    elif isinstance(o1, list):
        for i1, i2 in zip(o1, o2, strict=True):
            approx(i1, i2, rel=rel, abs=abs, nan_ok=nan_ok)
    elif isinstance(o1, dict):
        for k1, k2 in zip(sorted(o1), sorted(o2), strict=True):
            approx(o1[k1], o2[k2], rel=rel, abs=abs, nan_ok=nan_ok)
    else:
        assert o1 == pytest.approx(o2, rel=rel, abs=abs, nan_ok=nan_ok)



from typing import Optional

import pyudev


def uid_from_device(device: pyudev.Device) -> Optional[str]:
    parts = [
        device.get('ID_VENDOR_ID'),
        device.get('ID_MODEL_ID'),
        device.get('ID_SERIAL_SHORT') or device.get('ID_PATH') or device.get('DEVNAME'),
    ]
    return '-'.join(parts) if all(parts) else device.get('DEVNAME') or None


def scan_for_connected_devices() -> set[str]:
    devices = pyudev.Context().list_devices()
    video_device_ids = {uid_from_device(device) for device in devices if device.subsystem == 'video4linux'}
    return {uid for uid in video_device_ids if uid is not None}


def device_nodes_from_uid(uid: str) -> set[str]:
    devices = pyudev.Context().list_devices()
    matching_devices = [device for device in devices if uid_from_device(device) == uid]
    return {device.device_node for device in matching_devices if device.device_node is not None}



from __future__ import annotations

from typing import Optional

import numpy as np

from ..geometry import Pose, Spline
from ..helpers import angle
from .obstacle_map import ObstacleMap


class Path(list):

    @staticmethod
    def from_poses(poses) -> Path:
        steps = [Step(poses[0])]
        for pose in poses[1:]:
            steps += [Step(pose, steps[-1])]
        return Path(steps)

    def smooth(self, obstacle_map: ObstacleMap, control_dist: Optional[float] = None) -> None:
        while True:
            for s in range(1, len(self) - 1):
                if self[s].backward != self[s+1].backward:
                    continue

                new_step = Step(self[s+1].target, self[s].previous_step,
                                control_dist=control_dist, backward=self[s].backward)
                if not new_step.is_healthy():
                    continue
                new_distance = obstacle_map.get_minimum_spline_distance(new_step.spline, backward=new_step.backward)
                if new_distance == 0:
                    continue
                old_distance1 = obstacle_map.get_minimum_spline_distance(self[s].spline, self[s].backward)
                old_distance2 = obstacle_map.get_minimum_spline_distance(self[s+1].spline, self[s+1].backward)
                if new_distance < min(old_distance1, old_distance2):
                    continue
                self[s] = new_step
                del self[s+1]
                break  # restart while loop
            else:
                break  # exit while loop


class Step:

    def __init__(self,
                 target: tuple[float, float, float],
                 previous_step: Optional[Step] = None,
                 control_dist: Optional[float] = None,
                 backward: bool = False) -> None:
        self.previous_step = previous_step
        self.target = target
        self.backward = backward

        start = previous_step.target if previous_step is not None else target
        self.spline = Spline.from_poses(
            Pose(x=start[0], y=start[1], yaw=start[2]),
            Pose(x=target[0], y=target[1], yaw=target[2]),
            control_dist=control_dist, backward=backward)

    def __lt__(self, other: Step) -> bool:
        return self.target < other.target

    def is_healthy(self, curvature_limit=10.0) -> bool:
        dir_ = self.spline.start.direction(self.spline.end)
        yaw0 = self.spline.start.direction(self.spline.control1)
        yaw1 = self.spline.control2.direction(self.spline.end)
        return \
            np.abs(angle(dir_, yaw0)) < np.pi / 2 and \
            np.abs(angle(dir_, yaw1)) < np.pi / 2 and \
            np.abs(self.spline.max_curvature()) < curvature_limit  # NOTE: max_curvature can be NaN



from datetime import timedelta

from nicegui import ui

from . import rosys


class simulation_ui:

    def __init__(self) -> None:
        ui.label('Simulation speed')
        ui.slider(min=0, max=10, value=1, step=0.1, on_change=rosys.config.request_backup) \
            .bind_value(rosys.config, 'simulation_speed').props('label-always')
        self.simulation_time = ui.label()
        self.startup_time = rosys.time()
        ui.timer(0.1, self.update_simulation_time)

    def update_simulation_time(self) -> None:
        self.simulation_time.set_text(f'Running for {timedelta(seconds=int(rosys.time() - self.startup_time))}')



from __future__ import annotations

from nicegui import ui
from nicegui.elements.scene_objects import Extrusion, Group, Stl

from ..geometry import Prism
from ..rosys import config
from .odometer import Odometer


class RobotObject(Group):
    """The RobotObject UI element displays the robot with its given shape in a 3D scene.

    The current pose is taken from a given odometer.
    The `debug` argument can be set to show a wireframe instead of a closed polygon.
    """

    def __init__(self, shape: Prism, odometer: Odometer, *, debug: bool = False) -> None:
        super().__init__()
        self.shape = shape
        self.odometer = odometer
        self.robot_object: Extrusion | Stl
        with self:
            outline = [list(point) for point in self.shape.outline]
            self.robot_object = Extrusion(outline, self.shape.height, wireframe=debug)
            self.robot_object.material('#4488ff', 0.5)
        ui.timer(config.ui_update_interval, self.update)

    def with_stl(self, url: str, *,
                 x: float = 0, y: float = 0, z: float = 0,
                 omega: float = 0, phi: float = 0, kappa: float = 0,
                 scale: float = 1.0,
                 color: str = '#ffffff', opacity: float = 1.0) -> RobotObject:
        """Sets an STL to be displayed as the robot.

        The file can be served from a local directory with [app.add_static_files(url, path)](https://nicegui.io/reference#static_files).
        """
        self.robot_object.delete()
        with self:
            self.robot_object = Stl(url).move(x, y, z).rotate(omega, phi, kappa).scale(scale).material(color, opacity)
        return self

    def update(self) -> None:
        self.move(self.odometer.prediction.x, self.odometer.prediction.y)
        self.rotate(0, 0, self.odometer.prediction.yaw)



from dataclasses import dataclass

from ..geometry import Point, Pose


@dataclass(slots=True, kw_only=True)
class DelaunayPoseGroup:
    index: int
    point: Point
    neighbor_indices: list[int]
    poses: list[Pose]



import random
from typing import Optional

from typing_extensions import Self

from ... import rosys
from ..camera.configurable_camera import ConfigurableCamera
from ..image import Image, ImageSize
from .simulated_device import SimulatedDevice


class SimulatedCamera(ConfigurableCamera):

    def __init__(self,
                 *,
                 id: str,  # pylint: disable=redefined-builtin
                 name: Optional[str] = None,
                 connect_after_init: bool = True,
                 streaming: bool = True,
                 width: int = 800,
                 height: int = 600,
                 color: Optional[str] = None,
                 **kwargs,
                 ) -> None:
        super().__init__(id=id,
                         name=name,
                         connect_after_init=connect_after_init,
                         streaming=streaming,
                         **kwargs)
        self.device: Optional[SimulatedDevice] = None
        self.resolution = ImageSize(width=width, height=height)
        self._register_parameter('color', self.get_color, self.set_color,
                                 color or f'#{random.randint(0, 0xffffff):06x}')

    def to_dict(self) -> dict:
        return {
            'id': self.id,
            'width': self.resolution.width,
            'height': self.resolution.height,
        } | {
            name: param.value for name, param in self._parameters.items()
        }

    @classmethod
    def from_dict(cls, data: dict) -> Self:
        return cls(**data)

    @property
    def is_connected(self) -> bool:
        return self.device is not None

    async def connect(self) -> None:
        if not self.is_connected:
            self.device = SimulatedDevice(id=self.id, size=self.resolution)
            self._apply_all_parameters()

    async def disconnect(self) -> None:
        self.device = None

    async def capture_image(self) -> None:
        if not self.is_connected:
            return None
        image = Image(time=rosys.time(), camera_id=self.id, size=self.resolution)
        if rosys.is_test:
            image.data = b'test data'
        else:
            assert self.device is not None
            image.data = await rosys.run.cpu_bound(self.device.create_image_data)
        self._add_image(image)

    def set_color(self, val: str) -> None:
        assert self.device is not None
        self.device.color = val

    def get_color(self) -> Optional[str]:
        assert self.device is not None
        return self.device.color



from typing import Optional

from ...geometry import Rectangle
from ..image import ImageSize
from ..image_rotation import ImageRotation
from .camera import Camera


class TransformableCamera(Camera):

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.crop: Optional[Rectangle] = None
        """region to crop on the original resolution before rotation"""
        self.rotation: ImageRotation = ImageRotation.NONE
        """rotation which should be applied after grabbing and cropping"""

        self._resolution: Optional[ImageSize] = None

    def _resolution_after_transform(self, original_resolution: ImageSize) -> ImageSize:
        width = int(self.crop.width) if self.crop else original_resolution.width
        height = int(self.crop.height) if self.crop else original_resolution.height
        if self.rotation in {ImageRotation.LEFT, ImageRotation.RIGHT}:
            width, height = height, width
        return ImageSize(width=width, height=height)

    @property
    def rotation_angle(self) -> int:
        """Rotation angle in degrees."""
        return int(self.rotation)

    @rotation_angle.setter
    def rotation_angle(self, value: int) -> None:
        self.rotation = ImageRotation.from_degrees(value)

    def rotate_clockwise(self) -> None:
        """Rotate the image clockwise by 90 degrees."""
        self.rotation_angle += 90

    def rotate_counter_clockwise(self) -> None:
        """Rotate the image counter clockwise by 90 degrees."""
        self.rotation_angle -= 90



from dataclasses import dataclass

from ..geometry import Point


@dataclass(slots=True, kw_only=True)
class Polygon:
    outline: list[Point]

    def contains(self, point: Point) -> bool:
        inside = False
        for i, pi in enumerate(self.outline):
            pj = self.outline[(i + 1) % len(self.outline)]
            if ((pi.y > point.y) != (pj.y > point.y)) and (point.x < (pj.x - pi.x) * (point.y - pi.y) / (pj.y - pi.y) + pi.x):
                inside = not inside
        return inside



import asyncio
import logging
import time
from copy import copy
from multiprocessing import Pipe
from typing import Any

from .. import persistence, rosys, run
from ..driving import PathSegment
from ..event import Event
from ..geometry import Point, Pose, Prism, Spline
from .area import Area
from .obstacle import Obstacle
from .planner_process import (PlannerCommand, PlannerGrowMapCommand, PlannerObstacleDistanceCommand, PlannerProcess,
                              PlannerResponse, PlannerSearchCommand, PlannerTestCommand)


class PathPlanner(persistence.PersistentModule):
    """This module runs a path planning algorithm in a separate process.

    If given, the algorithm respects the given robot shape as well as a dictionary of accessible areas and a dictionary of obstacles, both of which a backed up and restored automatically.
    The path planner can search paths, check if a spline interferes with obstacles and get the distance of a pose to any obstacle.
    """

    def __init__(self, robot_shape: Prism) -> None:
        super().__init__()

        self.log = logging.getLogger('rosys.path_planner')

        self.connection, process_connection = Pipe()
        self.process = PlannerProcess(process_connection, robot_shape.outline)
        self.responses: dict[str, Any] = {}

        self.obstacles: dict[str, Obstacle] = {}
        self.areas: dict[str, Area] = {}

        self.OBSTACLES_CHANGED = Event()
        """the obstacles have changed (argument: dictionary of obstacles)"""
        self.AREAS_CHANGED = Event()
        """the areas have changed (argument: list of areas that have changed, can be None for all areas)"""

        rosys.on_startup(self.startup)
        rosys.on_shutdown(self.shutdown)
        rosys.on_repeat(self.step, 0.1)

    def backup(self) -> dict:
        finished_areas = {area_id: copy(area).close() for area_id, area in self.areas.items() if len(area.outline) >= 3}
        return {
            'obstacles': persistence.to_dict(self.obstacles),
            'areas': persistence.to_dict(finished_areas),
        }

    def restore(self, data: dict[str, Any]) -> None:
        persistence.replace_dict(self.obstacles, Obstacle, data.get('obstacles', {}))
        persistence.replace_dict(self.areas, Area, data.get('areas', {}))
        self.AREAS_CHANGED.emit(None)
        self.OBSTACLES_CHANGED.emit(self.obstacles)

    def startup(self) -> None:
        self.process.start()

    async def shutdown(self) -> None:
        self.log.info('stopping planner process...')
        self.connection.close()
        self.process.connection.close()
        self.process.join(5)
        if self.process.is_alive():
            self.process.terminate()
        else:
            if self.process.exitcode:
                self.log.info(f'bad exitcode for process: {self.process.exitcode}')
        self.log.info(f'teardown of {self.process} completed')

    async def step(self) -> None:
        try:
            if self.connection.poll():
                response = self.connection.recv()
                assert isinstance(response, PlannerResponse)
                if time.time() < response.deadline:
                    self.responses[response.id] = response.content
        except OSError as e:
            if 'handle is closed' in str(e):
                self.log.info('path planner process connection closed')
            else:
                raise

    async def grow_map(self, points: list[Point], timeout: float = 3.0) -> None:
        return await self._call(PlannerGrowMapCommand(
            points=points,
            deadline=time.time()+timeout,
        ))

    async def search(self, *, start: Pose, goal: Pose, timeout: float = 3.0) -> list[PathSegment]:
        return await self._call(PlannerSearchCommand(
            areas=list(self.areas.values()),
            obstacles=list(self.obstacles.values()),
            start=start,
            goal=goal,
            deadline=time.time()+timeout,
        ))

    async def test_spline(self, spline: Spline, timeout: float = 3.0) -> bool:
        return await self._call(PlannerTestCommand(
            areas=list(self.areas.values()),
            obstacles=list(self.obstacles.values()),
            spline=spline,
            deadline=time.time()+timeout,
        ))

    async def get_obstacle_distance(self, pose: Pose, timeout: float = 3.0) -> float:
        return await self._call(PlannerObstacleDistanceCommand(
            areas=list(self.areas.values()),
            obstacles=list(self.obstacles.values()),
            pose=pose,
            deadline=time.time()+timeout,
        ))

    async def _call(self, command: PlannerCommand, check_interval: float = 0.1) -> Any:
        with run.cpu():
            self.connection.send(command)
            while command.id not in self.responses:
                if time.time() > command.deadline:
                    raise TimeoutError(f'process call {command.id} did not respond in time')
                if rosys.is_test:
                    await self.step()  # NOTE: otherwise step() is not called while awaiting response
                await asyncio.sleep(check_interval)
            result = self.responses.pop(command.id)
            if isinstance(result, Exception):
                raise result
            return result



from dataclasses import dataclass
from typing import Any, Callable, Optional, overload

from .camera import Camera


@dataclass(slots=True, kw_only=True)
class ParameterInfo:
    name: str
    min: Optional[Any] = None
    max: Optional[Any] = None
    step: Optional[Any] = None


@dataclass(slots=True, kw_only=True)
class Parameter:
    """A camera parameter which can be adjusted.

    Value ranges can either be a list of options or a min/max value with an optional step size.
    """
    info: ParameterInfo
    value: Any
    setter: Callable
    getter: Callable


class ConfigurableCamera(Camera):
    """A generalized interface for adjusting camera parameters like exposure, brightness or fps."""
    IGNORE_NONE_VALUES = True

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self._parameters: dict[str, Parameter] = {}
        self._pending_operations: int = 0

    @overload
    def _register_parameter(self, name: str, getter: Callable[[], Any], setter: Callable[[Any], None], default_value: Any) -> None:
        ...

    @overload
    def _register_parameter(self, name: str, getter: Callable[[], Any], setter: Callable[[Any], None], default_value: Any,
                            min_value: Any, max_value: Any, step: Any) -> None:
        ...

    def _register_parameter(self, name: str, getter: Callable[[], Any], setter: Callable[[Any], None], default_value: Any,
                            min_value: Any = None, max_value: Any = None, step: Any = None) -> None:
        info = ParameterInfo(name=name, min=min_value, max=max_value, step=step)
        self._parameters[name] = Parameter(info=info, getter=getter, setter=setter, value=default_value)

    def _apply_parameters(self, new_values: dict[str, Any], force_set: bool = False) -> None:
        if not self.is_connected:
            return
        for name, value in new_values.items():
            if name not in self._parameters:
                raise ValueError(f'Cannot set unknown parameter "{name}"')
            if value is None and self.IGNORE_NONE_VALUES:
                continue
            if not force_set and value == self._parameters[name].value:
                continue
            self._parameters[name].setter(value)
        self._update_parameter_cache()

    def _apply_all_parameters(self) -> None:
        self._apply_parameters(self.parameters, force_set=True)

    def _update_parameter_cache(self) -> None:
        if not self.is_connected:
            return
        for param in self._parameters.values():
            val = param.getter()
            if val is None and self.IGNORE_NONE_VALUES:
                continue
            param.value = val

    def set_parameters(self, new_values: dict[str, Any]) -> None:
        self._apply_parameters(new_values)

    @property
    def parameters(self) -> dict[str, Any]:
        return {key: param.value for key, param in self._parameters.items()}

    def get_capabilities(self) -> list[ParameterInfo]:
        return [param.info for param in self._parameters.values()]



from __future__ import annotations

import logging
import re
from typing import Optional

import cv2

from ... import run
from .usb_camera_scanner import device_nodes_from_uid

MJPG = cv2.VideoWriter_fourcc(*'MJPG')


def find_video_id(camera_uid: str) -> Optional[int]:
    device_nodes = device_nodes_from_uid(camera_uid)
    video_ids = [
        int(node.strip().removeprefix('/dev/video'))
        for node in device_nodes
        if node.startswith('/dev/video')
    ]
    return min(video_ids) if video_ids else None


class UsbDevice:

    def __init__(self, video_id: int, capture: cv2.VideoCapture) -> None:
        self.video_id: int = video_id
        self.capture: cv2.VideoCapture = capture

        self.exposure_min: int = 0
        self.exposure_max: int = 0
        self.exposure_default: int = 0
        self.has_manual_exposure: bool = False
        self.video_formats: set[str] = set()

        self.set_video_format()

    @staticmethod
    def from_uid(camera_id: str) -> Optional[UsbDevice]:
        video_id = find_video_id(camera_id)
        if video_id is None:
            logging.error(f'Could not find video device for camera {camera_id}')
            return None

        capture = UsbDevice.create_capture(video_id)
        if capture is None:
            logging.error(f'Could not open video device {video_id}')
            return None

        return UsbDevice(video_id=video_id, capture=capture)

    @staticmethod
    def create_capture(index: int) -> Optional[cv2.VideoCapture]:
        capture = cv2.VideoCapture(index)
        if capture is None:
            return None
        capture.set(cv2.CAP_PROP_BUFFERSIZE, 1)
        if not capture.isOpened():
            capture.release()
            return None
        return capture

    async def load_value_ranges(self) -> None:
        output = await self.run_v4l('--all')
        if output is None:
            return
        match = re.search(r'exposure_absolute.*: min=(\d*).*max=(\d*).*default=(\d*).*', output)
        if match is not None:
            self.has_manual_exposure = True
            self.exposure_min = int(match.group(1))
            self.exposure_max = int(match.group(2))
            self.exposure_default = int(match.group(3))
        else:
            self.has_manual_exposure = False
        output = await self.run_v4l('--list-formats')
        matches = re.finditer(r"$.*'(.*)'.*", output)
        for m in matches:
            self.video_formats.add(m.group(1))

    async def run_v4l(self, *args) -> str:
        cmd = ['v4l2-ctl', '-d', str(self.video_id)]
        cmd.extend(args)
        return await run.sh(cmd)

    def set_video_format(self) -> None:
        if 'MJPG' in self.video_formats:
            # NOTE enforcing motion jpeg for now
            if self.capture.get(cv2.CAP_PROP_FOURCC) != MJPG:
                self.capture.set(cv2.CAP_PROP_FOURCC, MJPG)
            # NOTE disable video decoding (see https://stackoverflow.com/questions/62664621/read-jpeg-frame-from-mjpeg-self-without-decoding-in-python-opencv/70869738?noredirect=1#comment110818859_62664621)
            if self.capture.get(cv2.CAP_PROP_CONVERT_RGB) != 0:
                self.capture.set(cv2.CAP_PROP_CONVERT_RGB, 0)
            # NOTE make sure there is no lag (see https://stackoverflow.com/a/30032945/364388)
            if self.capture.get(cv2.CAP_PROP_BUFFERSIZE) != 1:
                self.capture.set(cv2.CAP_PROP_BUFFERSIZE, 1)
